# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Iterable
from typing import MutableMapping
from typing import TypeVar

import httpx
import pydantic
from aiopki.ext.jose import JWKS
from canonical.exceptions import ProgrammingError

from oauthx.lib.models import AuthorizationResponse
from oauthx.lib.models import ClientAuthorizationState
from oauthx.lib.models import Grant
from oauthx.lib.models import JWTBearerAssertion
from oauthx.lib.models import ObtainedGrant
from oauthx.lib.models import Provider
from oauthx.lib.protocols import IStorage
from oauthx.lib.types import GrantType
from oauthx.lib.types import PrompType
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ResponseType
from oauthx.server.protocols import IRequestSubject
from oauthx.server.protocols import IResourceOwner
from .clientkey import ClientKey
from .confidentialclient import ConfidentialClient
from .publicclient import PublicClient


T = TypeVar('T')


class Client(pydantic.RootModel[ConfidentialClient | PublicClient]):
    Key: ClassVar[type[ClientKey]] = ClientKey

    @property
    def id(self) -> ClientKey:
        return ClientKey(self.root.client_id)

    @property
    def pk(self) -> ClientKey:
        return ClientKey(self.root.client_id)

    @property
    def provider(self) -> Provider:
        return self.root.provider

    def allows_response_type(self, response_type: ResponseType) -> bool:
        return self.root.allows_response_type(response_type)

    def allows_scope(self, scope: set[str]) -> bool:
        return self.root.allows_scope(scope)

    def can_encrypt(self) -> bool:
        return self.root.can_encrypt()

    def can_grant(self, grant_type: GrantType | None) -> bool:
        """Return a boolean indicating if the client allows the given
        `grant_type`.
        """
        return self.root.can_grant(grant_type)

    def can_redirect(self, redirect_uri: RedirectURI | None) -> bool:
        return self.root.can_redirect(redirect_uri)

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        userinfo['azp'] = str(self.id)

    def get_default_redirect_uri(self) -> RedirectURI:
        return self.root.get_default_redirect_uri()

    def get_display_name(self) -> str:
        return self.root.get_display_name()

    def get_sector_identifier(self) -> str:
        return self.root.sector_identifier

    def has_redirection_endpoints(self) -> bool:
        return self.root.has_redirection_endpoints()

    def is_confidential(self) -> bool:
        return self.root.is_confidential()

    def is_public(self) -> bool:
        return isinstance(self.root, PublicClient)

    def must_push(self) -> bool:
        return False

    def resource_owner(self, subject: IRequestSubject) -> IResourceOwner.Key:
        return IResourceOwner.Key(str(self.id), str(subject.id)) # type: ignore

    def requires_state(self) -> bool:
        return False

    def update_key(self, update: Callable[[bytes], None]):
        return self.root.update_key(update)

    @functools.singledispatchmethod
    async def authenticate(self, *args: Any, **kwargs: Any) -> bool:
        """Authenticates the grant using the method specified by the
        client.
        """
        return await self.root.authenticate(*args, **kwargs)

    @functools.singledispatchmethod
    async def grant(
        self,
        params: AuthorizationResponse,
        *args: Any,
        **kwargs: Any
    ) -> ObtainedGrant:
        raise NotImplementedError(params)

    @grant.register
    async def authorization_code( # type: ignore
        self,
        params: AuthorizationResponse,
        state: ClientAuthorizationState,
        http: httpx.AsyncClient | None = None,
    ) -> ObtainedGrant:
        """Use the query parameters that the authorization server
        supplied to the client redirection endpoint to obtain
        an access token.
        """
        if isinstance(params, dict):
            params = AuthorizationResponse.model_validate(params)
        params.raise_for_status()
        await params.verify(self.provider, state)

        grant = Grant.model_validate({
            'grant_type': 'authorization_code',
            'code': params.code,
            'redirect_uri': state.redirect_uri
        })
        return await self.grant(grant, http=http)

    async def authorize(
        self,
        *,
        redirect_uri: str | None = None,
        scope: Iterable[str] | None = None,
        response_type: ResponseType | None = None,
        state_class: type[T] = ClientAuthorizationState,
        prompt: Iterable[PrompType] | None = None
    ) -> T:
        """Create an authorization request and return the URL to
        which a user-agent must be redirectred.
        """
        return await self.root.authorize( # type: ignore
            redirect_uri=redirect_uri,
            scope=scope,
            response_type=response_type,
            state_class=state_class,
            prompt=list(prompt or [])
        )

    async def client_credentials(
        self,
        scope: Iterable[str],
        http: httpx.AsyncClient | None = None,
    ) -> ObtainedGrant:
        grant = Grant.model_validate({
            'grant_type': 'client_credentials',
            'scope': scope
        })
        return await self.grant(grant, http=http)

    async def jwt(
        self,
        sub: str,
        authenticate: bool = False,
        http: httpx.AsyncClient | None = None,
        **claims: Any
    ):
        """Obtain an access token using a JWT bearer token."""
        if not self.is_confidential():
            raise ProgrammingError(
                "Only confidential clients can use the urn:ietf:para"
                "ms:oauth:grant-type:jwt-bearer grant. This client "
                f"uses: {type(self.root.credential).__name__}."
            )
        assert isinstance(self.root.credential, JWKS)
        assertion = await JWTBearerAssertion.new(
            signer=self.root.credential,
            iss=self.root.client_id,
            sub=sub,
            token_endpoint=str(self.provider.token_endpoint)
        )
        return await self.grant(assertion, authenticate=authenticate, http=http)

    async def persist(self, storage: IStorage) -> None:
        await storage.persist(self)

    @functools.singledispatchmethod
    async def refresh(
        self,
        obj: ObtainedGrant | str,
        *args: Any,
        **kwargs: Any
    ) -> ObtainedGrant:
        raise NotImplementedError
    
    @refresh.register
    async def _refresh_grant(
        self,
        grant: ObtainedGrant,
        *args: Any,
        **kwargs: Any
    ) -> ObtainedGrant:
        return await self.refresh(grant.refresh_token)

    @refresh.register
    async def _refresh_token(
        self,
        token: str,
        scope: Iterable[str] | None = None
    ) -> ObtainedGrant:
        grant = Grant.model_validate({
            'grant_type': 'refresh_token',
            'refresh_token': token,
            'scope': str.join(' ', list(sorted(filter(bool, set(scope))))) if scope else None
        })
        return await self.grant(grant)

    @grant.register
    async def _grant(
        self,
        grant: Grant,
        authenticate: bool = True,
        http: httpx.AsyncClient | None = None
    ) -> ObtainedGrant:
        if authenticate:
            await self.authenticate(grant)
        return await self.provider.grant(grant, http=http)

    @grant.register
    async def _(
        self,
        assertion: JWTBearerAssertion,
        authenticate: bool = True,
        http: httpx.AsyncClient | None = None
    ) -> ObtainedGrant:
        return await self.grant(
            assertion.model_dump(mode='json'),
            authenticate=authenticate,
            http=http
        )

    @grant.register
    async def _(
        self,
        params: dict, # type: ignore
        authenticate: bool = True,
        http: httpx.AsyncClient | None = None
    ) -> ObtainedGrant:
        return await self.grant(
            Grant.model_validate(params),
            authenticate=authenticate,
            http=http
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.server.models import ClientType
from oauthx.server.models import ClientKey
from oauthx.server.params import Storage
from oauthx.server.request import Request
from .query import CLIENT_ID


__all__: list[str] = ['Client']


async def get(
    request: Request,
    storage: Storage,
    client_id: str | None = CLIENT_ID
) -> ClientType | None:
    if client_id is None:
        return None
    client = await storage.get(ClientKey(client_id))
    if client is None:
        return None
    request.set_client(client)
    return client.root # type: ignore


Client: TypeAlias = Annotated[
    ClientType | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.params import ClientAuthorizationState
from oauthx.server.types import StopSnooping


__all__: list[str] = ['ReturnURL']


async def get(
    state: ClientAuthorizationState,
) -> str:
    if state is None:
        raise StopSnooping
    return_url = state.annotation('return-url')
    if return_url is None:
        raise NotImplementedError
    return return_url


ReturnURL: TypeAlias = Annotated[str, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from oauthx.lib.exceptions import UserAgentException
from oauthx.lib.types import RedirectURI


class AccountIssues(UserAgentException):
    __module__: str = 'oauthx.types'
    template_name: str = 'oauthx/errors/account.html.j2'
    status_code: int = 400
    
    def __init__(
        self,
        *,
        allow_redirect: bool = False,
        redirect_uri: RedirectURI | None = None,
        context: dict[str, Any] | None = None
    ):
        return_url = None
        if redirect_uri is not None and allow_redirect:
            return_url = redirect_uri.redirect(error='access_denied')
        super().__init__(
            error='access_denied',
            error_description="There are issues with your account. Please contact support.",
            context={
                **(context or {}),
                'allow_redirect': allow_redirect,
                'return_url': return_url
            }
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import TypeVar

from oauthx.lib.types import MaskedPrincipal


T = TypeVar('T')


class IPrincipalMasker:
    __module__: str = 'oauthx.server.types'

    async def mask(
        self,
        value: Any,
        factory: Callable[[str], T] = MaskedPrincipal
    ) -> T:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
from canonical.protocols import ITemplateService


__all__: list[str] = [
    'TemplateServiceDependency',
    'TemplateService',
]


def get(request: fastapi.Request) -> ITemplateService:
    return getattr(request, 'templates')



TemplateServiceDependency: ITemplateService = fastapi.Depends(get)
TemplateService: TypeAlias = Annotated[ITemplateService, TemplateServiceDependency]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
from aiopki.ext.jose import OIDCToken


class PersonalName(pydantic.BaseModel):
    name: str | None = None
    given_name: str | None = None
    family_name: str | None = None
    middle_name: str | None = None
    nickname: str | None = None

    @classmethod
    def from_oidc_token(cls, token: OIDCToken):
        return cls.model_validate(token.model_dump())

    def model_post_init(self, _: Any) -> None:
        if not self.is_provided():
            raise ValueError("None of the name elements are provided.")

    def is_provided(self):
        return any([
            self.name,
            self.given_name,
            self.family_name,
            self.middle_name,
            self.nickname
        ])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .application import Application
from .assertion import Assertion
from .authorization import Authorization
from .authorizationcode import AuthorizationCode
from .authorizationcontext import AuthorizationContext
from .authorizationrequest import AuthorizationRequest
from .authorizationstate import AuthorizationState
from .baseclient import BaseClient
from .claim import Claim
from .claimset import ClaimSet
from .client import Client
from .clientkey import ClientKey
from .confidentialclient import ConfidentialClient
from .grantedauthorizationcode import GrantedAuthorizationCode
from .issuedtoken import IssuedToken
from .objectfactory import ObjectFactory
from .principal import Principal
from .principal import PrincipalType
from .principal import PrincipalValueType
from .publicclient import PublicClient
from .receipt import Receipt
from .resourceowner import ResourceOwner
from .resourceownerkey import ResourceOwnerKey
from .responsemode import ResponseMode
from .scope import Scope
from .session import Session
from .subject import Subject
from .subjectkey import SubjectKey
from .userinfo import UserInfo


__all__: list[str] = [
    'Authorization',
    'Assertion',
    'AuthorizationCode',
    'AuthorizationContext',
    'AuthorizationRequest',
    'AuthorizationState',
    'Application',
    'BaseClient',
    'Claim',
    'ClaimSet',
    'Client',
    'ClientKey',
    'ClientType',
    'ConfidentialClient',
    'GrantedAuthorizationCode',
    'IssuedToken',
    'ObjectFactory',
    'Principal',
    'PrincipalType',
    'PrincipalValueType',
    'PublicClient',
    'Receipt',
    'ResourceOwner',
    'ResourceOwnerKey',
    'ResponseMode',
    'Scope',
    'Session',
    'SubjectKey',
    'Subject',
    'UserInfo',
]

ClientType = ConfidentialClient | PublicClient


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets
from typing import ClassVar
from typing import Literal

import pydantic
from canonical import ResourceIdentifier

from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.models import AuthorizationRequestParameters
from oauthx.lib.models import AuthorizationResponse
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ResponseType
from oauthx.server.protocols import IRequestSubject
from oauthx.server.protocols import IResourceOwner
from .client import Client
from .resourceowner import ResourceOwner


DEFAULT_RESPONSE_MODES: dict[ResponseType, str] = {
    'code': 'query'
}


class AuthorizationState(pydantic.BaseModel):
    default_response_modes: ClassVar[dict[ResponseType, str]] = {
        'code'  : 'query',
        'token' : 'fragment',
    }

    class KeyType(ResourceIdentifier[str, 'AuthorizationState']):
        client_id: str
        openapi_example: str = 'FOpou5oouxBHGGXBYI23R7vfEc3HcRjd4-_5EYYzsuQ'
        openapi_title: str = 'Request ID'

        def __init__(self, request_id: str):
            self.request_id = request_id

        def __str__(self) -> str:
            return self.request_id
        
        def __eq__(self, key: object) -> bool:
            return isinstance(key, type(self)) and key.request_id == self.request_id

    @property
    def client_id(self) -> Client.Key:
        return Client.Key(self.params.client_id)

    @property
    def owner(self) -> ResourceOwner.Key:
        assert self.client_id
        assert self.sub
        return ResourceOwner.Key(self.params.client_id, self.sub)

    @property
    def response_mode(self) -> str:
        return self.default_response_modes[self.response_type]

    @property
    def response_type(self) -> ResponseType:
        return self.params.response_type

    code: str | None = pydantic.Field(
        default=None
    )

    consumed: bool = pydantic.Field(
        default=False
    )

    created: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    id: str = pydantic.Field(
        default=...
    )

    params: AuthorizationRequestParameters = pydantic.Field(
        default=...
    )

    redirect_uri: RedirectURI | None = pydantic.Field(
        default=None
    )

    signed: bool = pydantic.Field(
        default=False
    )

    source: Literal['ENDPOINT', 'REMOTE', 'PUSHED'] = pydantic.Field(
        default=...
    )

    sub: str | None = pydantic.Field(
        default=None
    )

    async def get_redirect_uri(self) -> str:
        assert self.redirect_uri is not None
        params = AuthorizationResponse.model_validate({
            'code': self.code,
            'state': self.params.state
        })
        return self.redirect_uri.redirect(
            mode=self.response_mode,
            **params.root.model_dump(exclude_defaults=True, exclude_none=True)
        )

    def allows_redirect(self, redirect_uri: RedirectURI | None) -> bool:
        # If the redirect_uri was included in the authorization
        # request, these values must match.
        return self.params.redirect_uri == redirect_uri

    def authorize(self):
        """Authorizes the authorization request."""
        self.code = secrets.token_urlsafe(32)

    def authenticate(self, subject: IRequestSubject):
        """Authenticate the request. Raise an exception if the request is not
        valid.
        """
        if self.sub is None:
            self.sub = subject.id
        if self.sub != subject.id:
            raise InvalidRequest(
                "This authorization request can not be fulfilled."
            )
        
    def consume(self) -> None:
        self.consumed = True

    def has_consent(self, owner: IResourceOwner) -> bool:
        return True

    def is_authorized(self) -> bool:
        return self.code is not None

    def is_consumed(self) -> bool:
        return self.consumed

    def is_openid(self) -> bool:
        return 'openid' in self.params.scope



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.types import RedirectURI as RedirectURIType
from oauthx.server.request import Request
from oauthx.server.types import InvalidRedirectURI
from .authorizationrequest import AuthorizationRequest
from .client import Client
from .query import REDIRECT_URI


__all__: list[str] = [
    'RedirectURI'
]


async def get(
    request: Request,
    client: Client,
    params: AuthorizationRequest,
    redirect_uri: str | None = REDIRECT_URI,
) -> RedirectURIType | None:
    if params and redirect_uri:
        raise InvalidRequest(
            "The redirect_uri parameter must not be used in combination "
            "with request or request_uri"
        )
    if params is not None:
        redirect_uri = params.redirect_uri
    if redirect_uri is not None:
        try:
            redirect_uri = RedirectURIType(redirect_uri)

            # If there are parameters, the request was signed by the
            # client so we can trust the redirect URI.
            if not params and not client.can_redirect(redirect_uri):
                raise InvalidRedirectURI
        except ValueError:
            raise InvalidRedirectURI
    return redirect_uri


RedirectURI: TypeAlias = Annotated[
    RedirectURIType | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .iclient import IClient
from .iclientauthorizationstate import IClientAuthorizationState
from .iclientstorage import IClientStorage
from .basesettings import BaseSettings
from .iregisteredclient import IRegisteredClient
from .istorage import IStorage


__all__: list[str] = [
    'BaseSettings',
    'IClient',
    'IClientAuthorizationState',
    'IClientStorage',
    'IRegisteredClient',
    'IStorage'
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from gettext import gettext as _
from typing import Any

import fastapi

from oauthx.server.config import ProviderConfig
from oauthx.server.handlers import BaseLoginRequestHandler


class EnrolledLoginRequestHandler(BaseLoginRequestHandler):
    __module__: str = 'oauthx.ext.google'

    def get_context(self, request: fastapi.Request) -> dict[str, Any]:
        ctx = super().get_context(request)
        return {
            **ctx,
            'deny_url': None,
            'page_title': _("Sign in for work or school"),
            'previous_url': request.url_for('oauth2.login')\
                .include_query_params(**request.query_params),
        }

    def get_templates(self) -> list[str]:
        return ['oauthx/login.upstream.enrolled.html.j2']

    def get_providers(self) -> list[ProviderConfig]:
        return [x for x in self.config.providers if x.audience == 'institutional']


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pytest

from oauthx.client import  Client
from .types import AuthorizeFactory



@pytest.mark.skip("Needs implementation of SQL storage.")
@pytest.mark.asyncio
@pytest.mark.parametrize("scope", [
    {"email"}
])
async def test_userinfo_contains_scope_claims(
    authorize_request: AuthorizeFactory,
    agent: httpx.AsyncClient,
    client: Client,
    scope: set[str]
):
    _, response, state = await authorize_request(scope=scope)
    grant = await client.authorization_code(response, state, http=agent)
    response = await grant.access_token.userinfo(client, http=agent)
    assert hasattr(response, 'sub')


@pytest.mark.skip("Needs implementation of SQL storage.")
@pytest.mark.asyncio
async def test_userinfo_uses_get(
    authorize_request: AuthorizeFactory,
    agent: httpx.AsyncClient,
    client: Client
):
    _, response, state = await authorize_request()
    grant = await client.authorization_code(response, state, http=agent)
    await grant.access_token.userinfo(client, http=agent)


@pytest.mark.skip("Needs implementation of SQL storage.")
@pytest.mark.asyncio
async def test_userinfo_contains_sub(
    authorize_request: AuthorizeFactory,
    agent: httpx.AsyncClient,
    client: Client
):
    _, response, state = await authorize_request()
    grant = await client.authorization_code(response, state, http=agent)
    response = await grant.access_token.userinfo(client, http=agent)
    assert hasattr(response, 'sub')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import TypeVar

import fastapi
import pydantic

from .iclient import IClient
from .iclientauthorizationstate import IClientAuthorizationState
from .istorage import IStorage


P = TypeVar('P', bound=pydantic.BaseModel)
T = TypeVar('T')


class IClientStorage(IStorage, Protocol):
    __module__: str = 'oauthx.lib.protocols'

    @classmethod
    def inject(cls) -> Any:
        def f(request: fastapi.Request, storage: cls = fastapi.Depends(cls)):
            setattr(request, 'storage', storage)
        return fastapi.Depends(f)

    async def application(self, client_id: str) -> IClient | None: ...
    async def state(self, state: str) -> IClientAuthorizationState | None: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import asyncio.coroutines
import inspect
from typing import Annotated
from typing import Any
from typing import Callable
from typing import Literal

import fastapi

from oauthx.lib.utils import merge_signatures
from oauthx.lib.utils import merged_call
from .request import Request
from .params import ContentEncryptionKey
from .params import CurrentSubject
from .params import Storage
from .plugin import Plugin
from .types import AuthorizationRequestKey
from .types import StopSnooping


class PluginEndpoint:
    __module__: str = 'oauthx.server'
    _is_coroutine = asyncio.coroutines._is_coroutine # type: ignore

    @property
    def __signature__(self) -> inspect.Signature:
        sig = merge_signatures([
            inspect.signature(self.__call__),
            inspect.signature(self.plugin_class),
            inspect.signature(self.func)
        ])
        if self.context == 'authorize':
            sig = merge_signatures([sig, inspect.signature(self.authorization)])
        return sig
            

    def authorization(self, authorization_id: Annotated[AuthorizationRequestKey | None, fastapi.Path(..., alias='pk')]) -> None:
        pass

    def __init__(
        self,
        context: Literal['authorize', 'login', 'token'],
        plugin_class: type[Plugin],
        func: Callable[..., Any],
        authenticated: bool = True,
        needs_data: bool = False
    ) -> None:
        self.authenticated = authenticated
        self.context = context
        self.plugin_class = plugin_class
        self.needs_data = needs_data
        self.func = func

        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self) # nosec

    async def __call__(
        self,
        request: Request,
        key: ContentEncryptionKey,
        storage: Storage,
        subject: CurrentSubject,
        authorization_id: AuthorizationRequestKey | None,
        **kwargs: Any
    ) -> Any:
        if self.needs_data and subject is not None:
            await subject.decrypt_keys(key)

        authorization = None
        if authorization_id is not None:
            authorization = await storage.get(authorization_id)
            if authorization is None:
                raise StopSnooping
            request.set_context(await authorization.get_context(storage))
        return await merged_call(self.func, {
            'request': request,
            'storage': storage,
            'subject': subject,
            **kwargs,
            'self': merged_call(self.plugin_class, {
                **kwargs,
                'authorization': authorization,
                'request': request,
                'storage': storage,
                'subject': subject
            })
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

import pytest

from .types import AuthorizeFactory


@pytest.fixture
def response_mode() -> str:
    return 'query.jwt'


@pytest.mark.asyncio
async def test_jarm_response_contains_error_parameters(authorize_request: AuthorizeFactory):
    # Use an invalid scope to trigger an error.
    _, response, _ = await authorize_request(
        scope=bytes.hex(os.urandom(16)),
        expect_error='invalid_scope',
        expect_params={'response'}
    )
    assert response.is_error()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import AsyncGenerator
from typing import Annotated
from typing import TypeAlias

import fastapi
import httpx


__all__: list[str] = ['HTTPClientDependency']


async def get(request: fastapi.Request) -> AsyncGenerator[httpx.AsyncClient, None]:
    if getattr(request, 'http', None) is not None:
        yield getattr(request, 'http')
    else:
        client = httpx.AsyncClient()
        try:
            await client.__aenter__()
            yield client
        finally:
            await client.__aexit__(None, None, None)



HTTPClientDependency: httpx.AsyncClient = fastapi.Depends(get)
HTTPClient: TypeAlias = Annotated[httpx.AsyncClient, HTTPClientDependency]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets
from typing import Any
from typing import ClassVar
from typing import MutableMapping

import pydantic
from canonical import UnixTimestamp

from .sessionkey import SessionKey


class Session(pydantic.BaseModel):
    Key: ClassVar[type[SessionKey]] = SessionKey
    id: str
    sub: str | None = None
    email: str | None = None
    auth_time: UnixTimestamp | None = None
    _is_dirty: bool = pydantic.PrivateAttr(default=False)

    @classmethod
    def new(cls):
        return cls.model_validate({'id': secrets.token_urlsafe(64)})

    def authenticate(self, sub: str, email: str | None) -> None:
        """Authenticates the session as the given subject."""
        self.set('auth_time', UnixTimestamp.now(datetime.timezone.utc))
        self.set('sub', sub)
        if email is not None:
            self.set('email', email)

    def claims(self) -> dict[str, Any]:
        return self.model_dump(include={'id', 'sub', 'email', 'auth_time'})

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        if self.auth_time is not None:
            userinfo['auth_time'] = self.auth_time
        if self.email is not None:
            userinfo['email'] = self.email

    def is_dirty(self) -> bool:
        return self._is_dirty

    def set(self, key: str, value: Any) -> None:
        if key not in self.model_fields:
            raise AttributeError(key)
        if getattr(self, key) == value:
            return
        setattr(self, key, value)
        self._is_dirty = True
        


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal
from typing import Protocol

from aiopki.ext.jose import OIDCToken

from oauthx.lib.types import OIDCIssuerIdentifier
from .isubject import ISubject


class IAuthorizationServer(Protocol):
    """Specifies the interface for a concrete implementation of an
    OAuth 2.x/OpenID Connect authorization server.
    """
    __module__: str = 'oauthx.types'
    
    #: Raised when a set of principals resolves to multiple subjects.
    AmbiguousPrincipal: type[Exception] = type(
        'AmbiguousPrincipal', (Exception,), {}
    )

    #: Raised when we receive an assertion that is not trusted.
    UntrustedAssertion: type[Exception] = type(
        'UntrustedAssertion', (Exception,), {}
    )
    
    #: Raised when a Principal is orphaned.
    OrphanedPrincipal: type[Exception] = type(
        'OrphanedPrincipal', (Exception,), {}
    )

    async def onboard_oidc(
        self,
        token: OIDCToken,
        use: Literal['personal', 'institutional'],
        request_id: int | None = None,
        client_id: str | None = None
    ) -> tuple[ISubject, bool]:
        """Onboard or retrieve a :class:`~oauthx.types.ISubject` using
        an OpenID Connect ID Token. Return a tuple containing the
        :class:`ISubject` instance and a boolean indicating if the
        subject was onboarded (it did not previously exist).
        
        Args:
            token (aiopki.ext.jose.OIDCToken): the OpenID Connect ID Token
                that was obtained at a trusted identity provider.

        Returns:
            :class:`tuple`

        Raises:
            :exc:`AmbiguousPrincipal`: the principals
                included in the ID Token resolved to multiple
                subjects.
            :exc:`OrphanedPrincipal`: the principals provided were known
                by the authorization server but could not resolve to
                a subject.
            :exc:`UntrustedAssertion`: the ID Token did not include
                any principals that were trusted by the authorization
                server.
        """
        ...

    async def register_principal(
        self,
        *,
        issuer: OIDCIssuerIdentifier,
        subject: ISubject,
        value: Any,
        created: bool = False,
        verified: bool = False
    ) -> None:
        """Registers the given principals for the subject."""
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import uuid
from typing import Any

from canonical.protocols import ITemplateService
from fastapi.responses import HTMLResponse
from fastapi import Request
from fastapi import Response


class UserAgentException(Exception):
    """An exception class that indicates that a HTML response
    may be sent to the user agent.

    Args:
        error (str): the OAuth 2.x/OpenID Connect error code
            identifying the error condition that occurred. 
    """
    __module__: str = 'oauthx.types'
    allow_redirect: bool
    context: dict[str, Any]
    status_code: int = 400
    template_name: str

    def __init__(
        self,
        *,
        error: str,
        error_description: str | None = None,
        context: dict[str, Any] | None = None
    ):
        self.context = context or {}
        self.error = error
        self.error_description = error_description
        self.id = uuid.uuid4()

    async def render_to_response(
        self,
        request: Request,
        templates: ITemplateService
    ) -> Response:
        return HTMLResponse(
            status_code=self.status_code,
            headers={
                'X-Error': self.error,
                'X-Error-Description': self.error_description or ''
            },
            content=await templates.render_template(
                self.get_template_names(),
                context={
                    **self.context,
                    'error': self.error,
                    'error_description': self.error_description,
                    'exception': self,
                    'incident_id': self.id,
                    'request': request
                }
            )
        )
        
    def get_template_names(self) -> list[str] | str:
        return self.template_name


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Union

import pydantic
from aiopki.types import EncryptionResult
from aiopki.types import SubjectID
from canonical import EmailAddress
from canonical import Phonenumber

from oauthx.lib.types import MaskedPrincipal
from oauthx.server.protocols import ISubject
from .emailprincipal import EmailPrincipal
from .phonenumberprincipal import PhonenumberPrincipal
from .subjectkey import SubjectKey
from .subjectidentifierprincipal import SubjectIdentifierPrincipal


__all__: list[str] = [
    'PrincipalType',
    'PrincipalValueType',
    'Principal'
]


PrincipalType = Union[
    EmailPrincipal,
    PhonenumberPrincipal,
    SubjectIdentifierPrincipal
]


PrincipalValueType = Union[
    EmailAddress,
    Phonenumber,
    SubjectID
]


class Principal(pydantic.RootModel[PrincipalType]):

    @property
    def kind(self) -> str:
        return self.root.kind

    @property
    def masked(self) -> MaskedPrincipal:
        return self.root.masked

    @property
    def owner(self) -> SubjectKey:
        return SubjectKey(str(self.root.owner))

    @property
    def value(self) -> PrincipalValueType:
        assert not isinstance(self.root.value, EncryptionResult)
        return self.root.value

    def is_encrypted(self) -> bool:
        return isinstance(self.root.value, EncryptionResult)

    def is_owned_by(self, sub: SubjectKey) -> bool:
        """Return a boolean indicating if the subject identified by the
        key owns this principal.
        """
        return sub == self.owner

    def is_verified(self) -> bool:
        return self.root.verified

    async def encrypt(self, subject: ISubject) -> None:
        await self.root.encrypt(subject)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

import pydantic
from aiopki.ext.jose import CompactSerialized
from aiopki.ext.jose import JWT
from aiopki.lib import JSONWebKeySet
from aiopki.types import IDecrypter

from .error import Error
from .jarmauthorizationresponse import JARMAuthorizationResponse


class JARMToken(pydantic.BaseModel):
    response: CompactSerialized

    def is_valid(self, issuer: str, client_id: str, jwt: JWT) -> bool:
        return all([
            jwt.validate_iss(issuer),
            jwt.validate_aud(client_id),
            jwt.validate_exp(),
            jwt.validate_iat(max_age=300, required=False),
            jwt.validate_nbf(required=False),
        ])

    def raise_for_status(self):
        pass

    async def decrypt(self, decrypter: IDecrypter) -> None:
        if not decrypter.can_decrypt():
            raise TypeError(f'{type(decrypter).__name__} can not decrypt.')
        await self.response.decrypt(decrypter)

    async def decode(
        self,
        issuer: str,
        client_id: str,
        verifier: JSONWebKeySet,
        decrypter: IDecrypter
    ) -> JARMAuthorizationResponse | Error:
        if self.response.is_encrypted():
            await self.decrypt(decrypter)
        if not await self.response.verify(verifier):
            raise TypeError
        jwt = self.response.payload(factory=JWT.model_validate)
        if not self.is_valid(issuer, client_id, jwt):
            raise NotImplementedError
        if jwt.claims.get('error'):
            return Error.model_validate(jwt.model_dump())
        return JARMAuthorizationResponse.model_validate(jwt.model_dump())


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal
from typing import TypeAlias


GrantType: TypeAlias = Literal[
    'authorization_code',
    'client_credentials',
    'refresh_token',
    'urn:ietf:params:oauth:grant-type:jwt-bearer',
    'urn:ietf:params:oauth:grant-type:saml2-bearer',
    'urn:ietf:params:oauth:grant-type:token-exchange'
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any

import aiopki
import aiopki.types
import aiopki.ext.jose
import fastapi
import httpx
from aiopki.ext.keychain import Keychain
from canonical.protocols import ITemplateService

from oauthx.lib.protocols import IClient
from oauthx.lib.protocols import IStorage
from oauthx.lib.types import RedirectURI
from oauthx.server.protocols import IRequestSession
from oauthx.server.protocols import IRequestSubject
from .models import AuthorizationContext
from .models import ResponseMode


CACHE_HEADERS: set[str] = {
    'Cache-Control',
    'Expires',
    'Pragma',
}


class Request(fastapi.Request):
    context: AuthorizationContext | None = None
    http: httpx.AsyncClient | None = None
    keychain: Keychain
    oauth_client: IClient | None = None
    promptable: bool = False
    redirect_uri: RedirectURI | None = None
    response_mode: ResponseMode | None = None
    session_data: IRequestSession | None = None
    session_cookie: str
    signer: aiopki.types.ICryptoKey
    storage: IStorage
    subject: IRequestSubject | None = None
    templates: ITemplateService
    token_signer: aiopki.CryptoKeyType

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

    def authenticate(
        self,
        sub: str,
        email: str | None = None
    ) -> None:
        if self.session_data:
            self.session_data.authenticate(sub, email)

    def can_redirect(self) -> bool:
        return any([
            self.response_mode and self.response_mode.can_redirect(),
            self.context and self.context.can_redirect()
        ])

    def get_response_media_type(self, select: list[str]) -> str | None:
        accepts = {
            re.sub(';.*$', '', x)
            for x in str.split(self.headers.get('Accept', 'application/json'), ',')}
        return 'text/html'\
            if bool({'text/html', '*/*'} & accepts)\
            else 'application/json'

    def set_client(self, client: IClient):
        self.oauth_client = client

    def set_context(self, context: AuthorizationContext):
        self.context = context

    def set_session(self, session: IRequestSession) -> None:
        self.session_data = session

    async def error(
        self,
        template_names: str | list[str],
        error: str,
        error_description: str | None = None,
        error_uri: str | None = None,
        status_code: int = 400,
        media_type: str | None = None,
        context: dict[str, Any] | None = None,
        allow_redirect: bool = True
    ) -> fastapi.Response:
        accepts = self.get_response_media_type(['text/html', 'application/json']) or media_type
        context = {**(context or {}), 'error': error}
        headers: dict[str, str] = {
            'X-Error': error
        }
        if error_description:
            context['error_description'] = error_description
            headers['X-Error-Description'] = error_description
        if error_uri:
            context['error_uri'] = error_uri
            headers['X-Error-URI'] = error_uri
        if self.context:
            context.update(await self.context.get_template_context())
        match accepts:
            case 'text/html':
                if self.response_mode and self.response_mode.can_redirect():
                    response = await self.response_mode.redirect(
                        signer=self.token_signer,
                        error=context['error'],
                        error_description=context.get('error_description'),
                        error_uri=context.get('error_uri'),
                    )
                    response.headers.update(headers)
                elif self.context and allow_redirect and self.context.can_redirect():
                    response = fastapi.responses.RedirectResponse(
                        status_code=302,
                        headers=headers,
                        url=await self.context.error(
                            error=error,
                            error_description=error_description,
                            error_uri=error_uri
                        )
                    )
                else:
                    response = await self.render_to_response(
                        template_names=template_names,
                        status_code=status_code,
                        headers=headers,
                        context=context
                    )
            case 'application/json':
                response = fastapi.responses.JSONResponse(
                    status_code=status_code,
                    headers=headers,
                    content=context
                )
            case _:
                response = fastapi.responses.Response(
                    status_code=status_code,
                    headers=headers
                )
        return response

    @staticmethod
    async def sign_session(session: IRequestSession, signer: aiopki.types.ICryptoKey) -> str:
        jws = aiopki.ext.jose.jws(session.claims())
        await jws.sign(
            algorithm=signer.default_algorithm(),
            signer=signer,
            protected={'typ': 'jwt+session'}
        )
        return jws.encode(encoder=bytes.decode)

    async def process_response(self, response: fastapi.Response) -> fastapi.Response:
        # See https://curity.io/resources/learn/oauth-cookie-best-practices/
        if self.session_data and self.session_data.is_dirty():
            response.set_cookie(
                key=self.session_cookie,
                value=await self.sign_session(self.session_data, self.signer),
                max_age=(86400 * 365),
                secure=True,
                httponly=True,
                samesite='lax'
            )
        if not bool(set(response.headers) & CACHE_HEADERS):
            response.headers.update({
                'Cache-Control': 'no-cache, no-store, must-revalidate',
                'Expires': '0',
                'Pragma': 'Expires'
            })
        return response

    async def render_to_response(
        self,
        template_names: list[str] | str,
        context: dict[str, Any] | None = None,
        status_code: int = 200,
        headers: dict[str, Any] | None = None
    ) -> fastapi.Response:
        context = {
            **(context or {}),
            'client': self.oauth_client
        }
        return fastapi.responses.HTMLResponse(
            status_code=status_code,
            headers=headers,
            content=await self.templates.render_template(
                template_names,
                context=context or {}
            )
        )

    async def sign(
        self,
        message: bytes,
        algorithm: aiopki.types.IAlgorithm | None = None,
        using: str | None = None
    ) -> bytes:
        return await self.signer.sign(message, algorithm, using)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.types import ScopeType
from oauthx.server.types import InvalidAuthorizationRequest
from .authorizationcontext import AuthorizationContext
from .authorizationrequest import AuthorizationRequest
from .client import Client
from .query import SCOPE
from .redirecturi import RedirectURI
from .responsemode import ResponseMode


__all__: list[str] = [
    'Scope'
]


async def get(
    client: Client,
    context: AuthorizationContext,
    params: AuthorizationRequest,
    redirect_uri: RedirectURI,
    response_mode: ResponseMode,
    scope: str | None = SCOPE,
) -> ScopeType | None:
    requested = None
    if scope and params:
        raise InvalidRequest(
            "The scope parameter must not be included if the "
            "request or request_uri parameters are present."
        )
    if scope is not None:
        requested = ScopeType.fromstring(scope)
    if params:
        requested = params.scope
    if requested and not client.allows_scope(requested):
        raise InvalidAuthorizationRequest(
            error="invalid_scope",
            error_description="The client does not allow the requested scope.",
            context=context
        )
    return requested


Scope: TypeAlias = Annotated[
    ScopeType | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorization import AuthorizationRequestHandler
from .login import BaseLoginRequestHandler
from .login import NullLoginRequestHandler
from .token import TokenEndpointHandler
from .upstream import UpstreamBeginHandler
from .upstream import UpstreamCallbackHandler
from .userinfo import UserInfoEndpointHandler


__all__: list[str] = [
    'AuthorizationRequestHandler',
    'BaseLoginRequestHandler',
    'NullLoginRequestHandler',
    'TokenEndpointHandler',
    'UpstreamBeginHandler',
    'UpstreamCallbackHandler',
    'UserInfoEndpointHandler',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

import aiopki
import oauthx
from oauthx.lib import MemoryCache
from oauthx.lib.params import ApplicationCache
from oauthx.resource import ResourceRouter
from oauthx.resource.params import RequestSubject


aiopki.install_backend('aiopki.ext.azure')
aiopki.install_backend('aiopki.ext.google')


app: fastapi.FastAPI = oauthx.create_server(
    config='etc/oidc.conf',
    dependencies=[
        ApplicationCache(MemoryCache),
    ]
)


router = ResourceRouter(
    issuers={"https://checkpoint.localhost.webiam.id"},
    scope={'telegram'},
    audience='any'
)


@router.get('/')
async def f(subject: RequestSubject):
    return subject.model_dump()


app.include_router(router=router)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

import fastapi
import fastapi.params
from canonical.protocols import ICache

from oauthx.lib import MemoryCache
from oauthx.lib.params import ApplicationCache
from .models import *
from .params import *
from .resourceroute import ResourceRoute
from .resourcerouter import ResourceRouter


__all__: list[str] = [
    'AccessToken',
    'RequestAccessToken',
    'RequestSubject',
    'RequiredScope',
    'ResourceRoute',
    'ResourceRouter',
]


T = TypeVar('T', bound=fastapi.FastAPI)


def application_factory(
    cache: type[ICache] = MemoryCache,
    cls: type[T] = fastapi.FastAPI,
    **kwargs: Any
) -> T:
    dependencies: list[fastapi.params.Depends] = kwargs.setdefault('dependencies', [])
    dependencies.extend([
        ApplicationCache(cache)
    ])
    return cls(**kwargs)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import secrets
from typing import Any

import pydantic
from aiopki.ext.jose import JWA
from aiopki.ext.jose import JOSEObject
from aiopki.lib import JSONWebKeySet
from canonical import ResourceName
from fastapi.security import HTTPBasicCredentials

from oauthx.lib.models.assertion import Assertion
from .baseclient import BaseClient


class ConfidentialClient(BaseClient):
    credential: JSONWebKeySet | ResourceName | str = pydantic.Field(
        default=...,
        alias='client_secret'
    )
    
    # JWT Secured Authorization Response Mode for OAuth 2.0 (JARM)
    authorization_encrypted_response_alg: JWA | None = pydantic.Field(
        default=None
    )
    
    authorization_encrypted_response_enc: JWA | None = pydantic.Field(
        default=None
    )

    model_config = {
        'populate_by_name': True
    }

    def can_encrypt(self) -> bool:
        return isinstance(self.credential, JSONWebKeySet)

    def model_post_init(self, _: Any) -> None:
        if self.authorization_encrypted_response_enc\
        and not self.authorization_encrypted_response_alg:
            raise ValueError(
                "the authorization_encrypted_response_alg parameter "
                "is required."
            )
        if self.authorization_encrypted_response_alg\
        and not self.authorization_encrypted_response_enc:
            self.authorization_encrypted_response_enc = JWA.a128cbc_hs256

        # If the client wants to encrypt, then the credential must be
        # a JSONWebKeySet instance.
        if self.wants_encryption() and not isinstance(self.credential, JSONWebKeySet):
            raise ValueError(f"{type(self.credential)} can not encrypt.")

    def wants_encryption(self) -> bool:
        """Return a boolean indicating if the client implements encryption."""
        return any([
            self.authorization_encrypted_response_alg
        ])

    @functools.singledispatchmethod
    async def authenticate(
        self,
        _: Assertion | JOSEObject | HTTPBasicCredentials | str | None
    ) -> bool:
        return False

    @authenticate.register
    async def _(self, jws: JOSEObject) -> bool:
        return isinstance(self.credential, JSONWebKeySet) and await jws.verify(self.credential)

    @authenticate.register
    async def _(self, credential: str) -> bool:
        assert isinstance(self.credential, str)
        return secrets.compare_digest(credential, self.credential)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import uvicorn

from .ref import LoginRequestHandler
from .oidcrouter import OIDCRouter


app: fastapi.FastAPI = fastapi.FastAPI()
app.include_router(
    router=OIDCRouter.from_config(
        app=app,
        config='etc/oidc.conf',
        login_handler=LoginRequestHandler
    )
)


if __name__ == '__main__':
    uvicorn.run('oauthx.server.__main__:app', reload=True) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .jsonformatter import JSONFormatter
from .memorycache import MemoryCache
from .models import *
from .rfc9068accesstoken import RFC9068AccessToken


__all__: list[str] = [
    'Assertion',
    'AuthorizationCodeGrant',
    'AuthorizationRequestParameters',
    'AuthorizationResponse',
    'ClientAuthorizationState',
    'ClientCredentialsGrant',
    'Error',
    'Grant',
    'JARMToken',
    'JSONFormatter',
    'JWTBearerAssertion',
    'JWTBearerGrant',
    'MemoryCache',
    'ObtainedGrant',
    'Provider',
    'RedirectionParameters',
    'RFC9068AccessToken',
    'SAML2BearerAssertion',
    'ServerMetadata',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import StopSnooping
from oauthx.lib.models import ClientAuthorizationStateKey
from oauthx.lib.protocols import IClientAuthorizationState
from .query import STATE
from .storage import Storage


__all__: list[str] = ['ClientAuthorizationState']


async def get(
    storage: Storage,
    state: str | None = STATE
) -> IClientAuthorizationState | None:
    if state is None:
        return None
    obj = await storage.get(ClientAuthorizationStateKey(state))
    if obj is None:
        raise StopSnooping
    return obj


ClientAuthorizationState: TypeAlias = Annotated[
    IClientAuthorizationState | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from canonical.protocols import ICache


__all__: list[str] = ['ApplicationCache']


def ApplicationCache(cls: type[ICache]):
    async def get(
        request: fastapi.Request,
        cache: ICache = fastapi.Depends(cls)
    ) -> None:
        setattr(request, 'cache', cache)
    return fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Literal

import aiopki
import pydantic
from aiopki.ext.jose import JWKS
from aiopki.ext.jose import OIDCToken
from canonical.security import DomainACL

from oauthx.client.models import Client as ConsumingClient
from oauthx.lib.models import AuthorizationResponse
from oauthx.lib.models import ClientAuthorizationState
from oauthx.lib.models import ObtainedGrant
from oauthx.lib.types import ClientSecret
from oauthx.server.models import ClientType
from oauthx.server.types import UnusableAccount


class ProviderConfig(pydantic.BaseModel):
    audience: Literal['personal', 'institutional']
    client_id: str
    client: ConsumingClient | None = None
    credential: aiopki.CryptoKeyType | aiopki.StringSecret | ClientSecret
    display_name: str
    domains: DomainACL = pydantic.Field(
        default_factory=DomainACL.null
    )
    issuer: str
    logo_url: str
    name: str
    params: dict[str, str] = {}
    protocol: Literal['oidc']
    scope: list[str] = []
    trust_email: bool = False
    trust_phonenumber: bool = False

    def process_oidc_token(self, client: ClientType, state: ClientAuthorizationState, token: OIDCToken) -> OIDCToken:
        if not token.email:
            raise NotImplementedError
        token.email_verified = self.trust_email
        token.phone_number_verified = self.trust_phonenumber
        if not self.domains.allows(token.email.domain):
            raise UnusableAccount({
                'authorize_url': state.authorize_url,
                'allowed_domains': self.domains,
                'account_email': token.email,
                'account_email_domain': token.email.domain,
                'client_name': client.client_name,
                'provider': self,
                'reason': 'UNUSABLE_PROVIDER',
                'return_url': state.annotation('return-url'),
                'token': token
            })
        return token

    async def identify(
        self,
        *,
        client_id: str,
        redirect_uri: str,
        return_url: str,
        state_class: type[ClientAuthorizationState] = ClientAuthorizationState
    ) -> ClientAuthorizationState:
        await self
        assert self.client is not None
        return await self.client.authorize(
            redirect_uri=redirect_uri,
            scope=set(self.scope),
            response_type='code',
            state_class=state_class,
            prompt={'consent'},
            annotations={
                'client-id': client_id,
                'return-url': return_url,
                'provider': self.name
            },
            **self.params
        )

    async def discover(self):
        if self.client is None:
            if inspect.isawaitable(self.credential):
                self.credential = await self.credential
            if not isinstance(self.credential, (JWKS, str)):
                raise RuntimeError(
                    f"Unable to discover {type(self).__name__}.credential by "
                    f"awaiting, got {type(self.credential).__name__}"
                )
            self.client = ConsumingClient.model_validate({
                'provider': self.issuer,
                'client_id': self.client_id,
                'credential': self.credential
            })
            
    async def authorization_code(
        self,
        params: AuthorizationResponse,
        state: ClientAuthorizationState
    ) -> ObtainedGrant:
        await self
        assert self.client is not None
        obtained = await self.client.authorization_code(params, state)
        if obtained.id_token is None:
            raise NotImplementedError
        return obtained

    def __await__(self):
        return self.discover().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import dataclasses
import datetime
import hmac
from typing import Annotated
from typing import TypeAlias

import fastapi
from aiopki.types import SubjectID
from canonical import UnixTimestamp

from oauthx.lib.types import MaskedPrincipal
from oauthx.lib.types import OIDCIssuerIdentifier
from oauthx.server.params import Masker
from oauthx.server.types import StopSnooping
from ..const import ISSUER
from .telegramtoken import TelegramToken


__all__: list[str] = ['TelegramUserInfo']


MAX_AGE: int = 2


@dataclasses.dataclass
class UserInfo:
    id: SubjectID
    issuer: OIDCIssuerIdentifier
    username: str
    masked: MaskedPrincipal


async def get(
    request: fastapi.Request,
    masker: Masker,
    token: TelegramToken,
    auth_date: UnixTimestamp,
    user_id: int = fastapi.Query(alias='id'),
    username: str = fastapi.Query(),
    mac: str = fastapi.Query(alias='hash'),
) -> UserInfo:
    now = datetime.datetime.now(datetime.timezone.utc)
    age = abs(int((now - auth_date).total_seconds()))
    check_sequence = bytes.join(b'\n', [
        str.encode(f'{k}={v}') for k, v
        in sorted(request.query_params.items(), key=lambda x: x[0])
        if k != 'hash'
    ])
    is_valid = hmac.compare_digest(
        hmac.new(token, check_sequence, 'sha256').hexdigest(),
        mac
    )
    if not is_valid:
        raise StopSnooping
    if age > MAX_AGE:
        raise StopSnooping

    subject_id = SubjectID(iss=ISSUER, sub=str(user_id))

    return UserInfo(
        id=subject_id,
        issuer=OIDCIssuerIdentifier(ISSUER),
        masked=await masker.mask(subject_id),
        username=username
    )


TelegramUserInfo: TypeAlias = Annotated[UserInfo, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
import fastapi.params
from canonical.exceptions import ProgrammingError

from oauthx.server.types import StopSnooping
from oauthx.lib.params import Storage
from oauthx.server.models import ClientKey
from oauthx.server.models import ClientType
from oauthx.server.params import ClientAuthorizationState
from oauthx.server.request import Request


__all__: list[str] = [
    'ClientDependency',
    'Client'
]


async def get(
    request: Request,
    state: ClientAuthorizationState,
    storage: Storage
) -> ClientType:
    if state is None:
        raise StopSnooping
    client_id = state.annotation('client-id', decoder=ClientKey)
    if client_id is None:
        raise ProgrammingError(
            "The ClientAuthorizationState instance must be annotated with "
            "the 'client-id' annotation, point to an existing OAuth 2.x/"
            "OpenID Connect client."
        )
    client = await storage.get(client_id)
    if client is None:
        raise StopSnooping
    request.set_client(client.root) # type: ignore
    return client.root # type: ignore


ClientDependency: fastapi.params.Depends = fastapi.Depends(get)
Client: TypeAlias = Annotated[ClientType, ClientDependency]



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .clientstorage import ClientStorage
from .datastorestorage import DatastoreStorage
from .enrolledlogin import EnrolledLoginRequestHandler
from .login import LoginRequestHandler
from .storage import DatastoreStorage as DatastoreServerStorage


__all__: list[str] = [
    'ClientStorage',
    'DatastoreStorage',
    'DatastoreServerStorage',
    'EnrolledLoginRequestHandler',
    'LoginRequestHandler',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

import pytest
from aiopki.lib import JSONWebKey
from aiopki.utils import b64encode

from aiopki.lib import JSONWebKey
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key




@pytest.fixture(scope='function')
def random_signer() -> JSONWebKey:
    return JSONWebKey.model_validate({
        'kty': 'oct',
        'use': 'sig',
        'alg': 'HS384',
        'k': b64encode(os.urandom(32))
    })


@pytest.fixture(scope='session')
def rsa_key() -> RSAPrivateKey:
    return generate_private_key(65537, 2048)


@pytest.fixture(scope='session')
def rsa_enc(rsa_key: RSAPrivateKey) -> JSONWebKey:
    return JSONWebKey.model_validate({
        'use': 'enc',
        'key': rsa_key,
    })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from aiopki.types import EncryptionResult
from canonical import EmailAddress

from .baseclaim import BaseClaim


class EmailClaim(BaseClaim):
    kind: Literal['email'] = 'email'
    value: EmailAddress | EncryptionResult

    def get_encryption_input(self) -> bytes:
        assert isinstance(self.value, str)
        return str.encode(self.value, 'ascii')

    @BaseClaim.requires_decryption
    def get_masking_input(self) -> bytes:
        assert isinstance(self.value, EmailAddress)
        return self.value.maskable


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
from canonical import PythonSymbol

from oauthx.server.models import Scope


class ScopeSpecification(pydantic.BaseModel):
    name: str
    claims: list[str]
    impl: PythonSymbol[type[Scope]] = pydantic.Field(
        default_factory=lambda: PythonSymbol.fromqualname('oauthx.server.models.Scope') # type: ignore
    )
    
    @property
    def scope(self) -> Any:
        return self.impl.value(**self.model_dump(exclude={'impl'}))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .iresourceowner import IResourceOwner


class NullResourceOwner(IResourceOwner):
    __module__: str = 'oauthx.types'

    @property
    def pk(self) -> IResourceOwner.Key:
        raise NotImplementedError
    



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from oauthx.server.params import CurrentSubject
from oauthx.server.request import Request
from ..baserequesthandler import BaseRequestHandler


class UpstreamBeginHandler(BaseRequestHandler):
    __module__: str = 'oauthx.server.handlers'
    include_in_schema: bool = False
    name: str = 'oauth2.upstream.begin'
    methods: list[str] = ['POST']
    path: str = '/begin'
    response_description: str = "Initiate the authentication flow"
    subject: CurrentSubject
    summary: str = "Upstream provider"

    def setup(
        self,
        *,
        request_id: str = fastapi.Form(...),
        client_id: str = fastapi.Form(...),
        provider: str = fastapi.Form(...),
        next_url: str = fastapi.Form(...),
        subject: CurrentSubject,
        **_: Any
    ):
        self.client_id = client_id
        self.next_url = next_url
        self.request_id = request_id
        self.provider = self.config.get_provider(provider)
        self.subject = subject

    async def begin(self, request: Request) -> None:
        if self.provider is not None:
            await self.provider
        return await super().begin(request)

    async def handle(self, request: Request) -> fastapi.Response:
        if self.provider is None or self.subject is not None:
            return fastapi.Response(status_code=403)
        state = await self.provider.identify(
            client_id=self.client_id,
            redirect_uri=str(request.url_for('oauth2.upstream.callback')),
            return_url=self.next_url
        )
        state.annotate('request', str(self.request_id))
        await self.storage.persist(state)
        self.logger.debug(
            "Initiating OIDC login flow (provider: %s, client: %s, request: %s)",
            self.provider.name, self.client_id, self.request_id
        )
        return fastapi.responses.RedirectResponse(
            status_code=302,
            url=state.authorize_url
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.types import ResponseType as ResponseType_
from .authorizationrequest import AuthorizationRequest
from .query import RESPONSE_TYPE


__all__: list[str] = [
    'ResponseType'
]


async def get(
    params: AuthorizationRequest,
    response_type: ResponseType_ | None = RESPONSE_TYPE,
) -> ResponseType_ | None:
    if params and response_type:
        raise InvalidRequest(
            "The response_type parameter must not be included if the "
            "request or request_uri parameters are present."
        )
    if params:
        response_type = params.response_type
    return response_type


ResponseType: TypeAlias = Annotated[
    ResponseType_ | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
import pydantic
from aiopki.ext.jose import CompactSerialized

from oauthx.lib.exceptions import InvalidGrant
from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.types import GrantType
from oauthx.server.models import AuthorizationCode
from oauthx.server.params import Storage
from oauthx.server.protocols import IResourceOwner
from .query import CODE
from .query import GRANT_TYPE


__all__: list[str] = [
    'ResourceOwner'
]


def requires_owner(grant_type: str):
    return grant_type not in {
        'client_credentials',
        'urn:ietf:params:oauth:grant-type:jwt-bearer'
    }


async def get(
    storage: Storage,
    code: CompactSerialized | None = CODE,
    grant_type: GrantType = GRANT_TYPE
) -> IResourceOwner | None:
    owner: IResourceOwner | None
    match grant_type:
        case 'authorization_code':
            if code is None:
                raise InvalidRequest(
                    "The authorization_code parameter is required."
                )
            try:
                jwt = code.payload(AuthorizationCode.model_validate_json)
            except pydantic.ValidationError:
                # The code was signed by us but its structure did not validate.
                raise InvalidGrant(
                    "The provided authorization code is not valid."
                )
            owner = await storage.get(jwt.owner)
        case 'client_credentials':
            owner = None
        case 'urn:ietf:params:oauth:grant-type:jwt-bearer':
            # The primary key of a resource owner is (client_id, subject_id)
            # but for this grant the client might not identify itself. If
            # there is a resource owner will be determined after validating
            # the assertion.
            owner = None
        case _:
            raise NotImplementedError

    if requires_owner(grant_type) and owner is None:
        raise InvalidGrant(
            "The resource owner identified by the grant parameters "
            "does not exist."
        )
    return owner

ResourceOwner: TypeAlias = Annotated[
    IResourceOwner | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Callable
from typing import Protocol

from canonical import ResourceIdentifier


class IResourceOwner(Protocol):

    class Key(ResourceIdentifier[str, 'IResourceOwner']):
        client_id: str
        openapi_example: str = 'clients/client-123/resourceOwners/123'
        openapi_title: str = 'Resource Owner ID'

        def __init__(self, client_id: str, sub: str):
            self.client_id = client_id
            self.sub = sub
        
        def __eq__(self, key: object) -> bool:
            return isinstance(key, type(self)) and all([
                self.client_id == key.client_id,
                self.sub == key.sub
            ])

        def __str__(self) -> str:
            return f'{self.client_id}/{self.sub}'
        
        def __hash__(self) -> int:
            return hash((self.client_id, self.sub))
        
    @property
    def pk(self) -> Key: ...

    def grants_consent(self, scope: set[str]) -> bool: ...
    def update_key(self, update: Callable[[bytes], None]): ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
from typing import Any
from typing import Iterable

from oauthx.lib.types import MaskedPrincipal
from .models import PrincipalValueType
from .models import Principal
from .params import Masker
from .params import Storage


class SubjectResolver:
    """Provides an interface to resolve :class:`~oauthx.types.ISubject`
    instances.
    """
    __module__: str = 'oauthx.server'
    storage: Storage

    def __init__(
        self,
        storage: Storage,
        masker: Masker
    ):
        self.masker = masker
        self.storage = storage

    async def mask(self, value: PrincipalValueType | None) -> MaskedPrincipal | None:
        if value is None:
            return None
        return await self.masker.mask(value)

    async def resolve_principal(self, principal: Any) -> Principal | None:
        masked = await self.mask(principal)
        if masked is None:
            return None
        return await self.storage.get(masked)

    async def resolve_principals(
        self,
        principals: Iterable[Any]
    ) -> tuple[Principal | None, ...]:
        """Resolve the given iterable `principals` to :class:`Principal`
        instances.
        """
        principals = await asyncio.gather(*map(self.mask, principals))
        return tuple(await asyncio.gather(*map(self.storage.get, principals)))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any, Iterator
from typing import MutableMapping

from oauthx.lib.protocols import IStorage
from oauthx.server.protocols import ISubject
from .claim import Claim


class UserInfo(MutableMapping[str, Any]):
    _known: set[tuple[str, str]]
    claims: dict[str, Any]

    @classmethod
    async def restore(cls, subject: ISubject, storage: IStorage):
        q = storage.find(
            kind=Claim,
            filters=[('sub', '=', int(subject.get_primary_key()))], # type: ignore
            sort=['obtained']
        )
        self = cls()
        async for claim in q:
            await claim.decrypt(subject)
            claim.contribute_to_userinfo(self)
        return self

    def __init__(self):
        self.claims = {}

    def dump(self, claims: set[str]) -> dict[str, Any]:
        return {
            k: self.claims[k] for k in claims
            if self.claims.get(k) not in {None, ''}
        }

    def has(self, claim: str) -> bool:
        return claim in self.claims

    def __delitem__(self, __key: str) -> None:
        return self.claims.__delitem__(__key)

    def __getitem__(self, __key: str) -> Any:
        return self.claims.__getitem__(__key)

    def __iter__(self) -> Iterator[str]:
        return self.claims.__iter__()

    def __len__(self) -> int:
        return self.claims.__len__()

    def __setitem__(self, __key: str, __value: Any) -> None:
        self.claims[__key] = __value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .error import Error
from .fatalerror import FatalError
from .insufficientscope import InsufficientScope
from .interactionexception import InteractionException
from .invalidauthorizationresponse import InvalidAuthorizationResponse
from .invalidgrant import InvalidGrant
from .invalidredirecturi import InvalidRedirectURI
from .invalidrequest import InvalidRequest
from .invalidscope import InvalidScope
from .invalidtoken import InvalidToken
from .invalidtokenresponse import InvalidTokenResponse
from .stopsnooping import StopSnooping
from .trustissues import TrustIssues
from .unknownclient import UnknownClient
from .unsupportedresponsetype import UnsupportedResponseType
from .useragentexception import UserAgentException


__all__: list[str] = [
    'Error',
    'FatalError',
    'InsufficientScope',
    'InteractionException',
    'InvalidAuthorizationResponse',
    'InvalidGrant',
    'InvalidRedirectURI',
    'InvalidRequest',
    'InvalidScope',
    'InvalidToken',
    'InvalidTokenResponse',
    'StopSnooping',
    'TrustIssues',
    'UnknownClient',
    'UnsupportedResponseType',
    'UserAgentException',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from oauthx.lib.exceptions import InteractionException


class UnusableAccount(InteractionException):
    __module__: str = 'oauthx.types'
    allow_redirect: bool = False
    error_description: str = (
        "The account that was used to sign in is not usable."
    )
    status_code: int = 400
    template_name: str = 'oauthx/errors/login.unusable.html.j2'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .assertion import Assertion
from .client import Client
from .grant import Grant
from .authorization import Authorization
from .granttype import GrantType
from .query import REDIRECT_URI
from .resourceowner import ResourceOwner


__all__: list[str] = [
    'Assertion',
    'Authorization',
    'Client',
    'Grant',
    'GrantType',
    'ResourceOwner',
    'REDIRECT_URI',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from oauthx.lib.types import RedirectURI
from .basegrant import BaseGrant


class AuthorizationCodeGrant(BaseGrant):
    grant_type: Literal['authorization_code'] = pydantic.Field(
        default=...,
        description="This value **must** be `authorization_code`."
    )

    code: str = pydantic.Field(
        default=...,
        description="The authorization code received from the authorization server"
    )

    redirect_uri: RedirectURI | None = pydantic.Field(
        default=None,
        description=(
            "This parameter is **required** if the `redirect_uri` parameter was "
            "included in the authorization request and their values **must** be identical."
        )
    )

    @property
    def scope(self) -> set[str]:
        return set()

    def requires_offline_access(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Protocol


class IAuthorizationRouter(Protocol):
    __module__: str = 'oauthx.server.types'
    
    def add_plugin(
        self,
        *,
        methods: list[str],
        path: str,
        handler: type,
        method: Callable[..., Any],
        name: str,
        description: str,
        authenticated: bool = True,
        needs_data: bool = False
    ) -> None:
        ...

    def register_template_module(self, qualname: str) -> None:
        ...

    def register_scope(
        self,
       name: str,
       cls: type
    ) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from oauthx.lib.protocols import IClient
from .models import Authorization
from .params import ObjectFactory
from .params import Storage
from .params import CurrentSubject
from .request import Request


class Plugin:
    __module__: str = 'oauthx.server'

    def __init__(
        self,
        request: Request,
        factory: ObjectFactory,
        storage: Storage,
        subject: CurrentSubject,
        authorization: Authorization | None = None
    ) -> None:
        self.authorization = authorization
        self.factory = factory
        self.request = request
        self.storage = storage
        self.subject = subject

    @classmethod
    def __register__(cls, app: fastapi.FastAPI | fastapi.APIRouter) -> None:
        pass

    def handles_scope(self, name: str) -> bool:
        """Return a boolean indicating if the handler knows the
        give scope `name`.
        """
        return False

    async def render_to_response(
        self,
        template_names: list[str] | str,
        context: dict[str, Any] | None = None,
        status_code: int = 200,
        headers: dict[str, Any] | None = None
    ) -> fastapi.Response:
        return await self.request.render_to_response(
            template_names=template_names,
            context=context,
            status_code=status_code,
            headers=headers
        )

    async def resolve_scope(
        self,
        client: IClient,
        authorization: Authorization,
        name: str
    ) -> tuple[str | None, dict[str, str]]:
        """Ensure that the given scope can be granted. Return
        ``None`` if the scope can be granted, otherwise
        return a tuple containing a named URL (string) and
        its parameters (dict).
        """
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import UnknownClient
from oauthx.lib.protocols import IClient
from oauthx.server.params import Storage
from .assertion import Assertion
from .clientcredentials import ClientCredentials


__all__: list[str] = ['Client']


async def authenticate_client(
    *,
    storage: Storage,
    assertion: Assertion,
    credentials: ClientCredentials,
) -> IClient | None:
    client = None
    if assertion and assertion.is_self_issued():
        raise NotImplementedError
    if credentials and assertion:
        raise NotImplementedError
    if not credentials and not assertion:
        return None

    if credentials and credentials.client_secret:
        client = await storage.get(credentials.client_id)
        if client and not await client.authenticate(credentials.client_secret):
            return None

    if assertion and assertion.is_client():
        client = await storage.get(assertion.client_id)
        if client and not await client.authenticate(assertion.jws):
            return None

    return client


async def get(
    storage: Storage,
    assertion: Assertion,
    credentials: ClientCredentials,
) -> IClient | None:
    client: IClient | None = None
    has_credentials = any([
        credentials and credentials.client_secret,
        assertion
    ])
    if has_credentials:
        client = await authenticate_client(
            storage=storage,
            assertion=assertion,
            credentials=credentials,
        )
    else:
        if credentials is not None:
            client = await storage.get(credentials.client_id)
            if client and client.is_confidential():
                raise UnknownClient(
                    "Client authentication was not successful."
                )

    return client


Client: TypeAlias = Annotated[
    IClient | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import functools
import hashlib
import operator
import secrets
from typing import Any
from typing import Literal
from typing import Iterable

import fastapi
from aiopki.utils import b64encode
from aiopki.ext.jose import OIDCToken

from oauthx.lib.protocols import IStorage
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ResponseType
from oauthx.lib.types import OIDCIssuerIdentifier
from oauthx.server.models import BaseClient
from oauthx.server.protocols import IClient
from oauthx.server.protocols import ISubject
from oauthx.server.protocols import IUserInfoContributor
from oauthx.server.types import IPrincipalMasker
from oauthx.server.types import IResponseMode
from .authorization import Authorization
from .authorizationrequest import AuthorizationRequest
from .claim import Claim
from .principal import Principal
from .principal import PrincipalValueType
from .receipt import Receipt
from .responsemode import ResponseMode
from .scope import Scope
from .session import Session
from .subject import Subject
from .subjectkey import SubjectKey
from .userinfo import UserInfo


class ObjectFactory:
    """Provides an interface to create model instances used by
    the authorization server.
    """
    __module__: str = 'oauthx.server'
    default_scopes: list[Scope] = [
        Scope(name='openid', claims=[
            'acr',
            'amr',
            'auth_time',
            'azp',
            'nonce',
            'sub',
        ]),
        Scope(name='address', claims=['address']),
        Scope(
            name='email',
            claims=[
                'email',
                'email_verified'
            ]
        ),
        Scope(
            name='profile',
            claims=[
                'name',
                'family_name',
                'given_name',
                'middle_name',
                'nickname',
                'preferred_username',
                'profile',
                'picture',
                'website',
                'gender',
                'birthdate',
                'zoneinfo',
                'locale',
                'updated_at' 
            ]
        ),
        Scope(
            name='phone',
            claims=[
                'phone_number',
                'phone_number_verified'
            ]
        ),
    ]
    masker: IPrincipalMasker
    storage: IStorage

    def __init__(
        self,
        issuer: str,
        storage: IStorage,
        masker: IPrincipalMasker,
        scopes: list[Scope] | None = None
    ):
        self.scopes = {
            **{x.name: x for x in self.default_scopes},
            **{x.name: x for x in (scopes or [])}
        }
        self.issuer = issuer
        self.masker = masker
        self.storage = storage

    async def authorization(
        self,
        request: AuthorizationRequest,
        client_id: str,
        sub: SubjectKey,
        token_types: Iterable[str],
        lifecycle: Literal['GRANTED', 'ISSUED'] = 'ISSUED',
        scope: set[str] | None = None,
    ) -> Authorization:
        return Authorization.model_validate({
            'issuer': self.issuer,
            'client_id': str(client_id),
            'id': request.id,
            'lifecycle': lifecycle,
            'scope': scope or set(),
            'sub': sub,
            'token_types': set(token_types),
        })

    async def claim(
        self,
        receipt_id: int,
        kind: Literal['email', 'gender', 'name'],
        provider: str,
        issuer: str,
        sub: int,
        value: Any,
        now: datetime.datetime | None = None,
        ial: int = 0,
    ) -> Claim:
        now = now or datetime.datetime.now(datetime.timezone.utc)
        return Claim.new(
            receipt_id=receipt_id,
            id=await self.storage.allocate_identifier(Claim),
            kind=kind,
            provider=provider,
            issuer=issuer,
            sub=sub,
            value=value,
            now=now
        )

    async def principal(
        self,
        *,
        subject: ISubject,
        issuer: Literal['self'] | OIDCIssuerIdentifier,
        owner: SubjectKey,
        value: PrincipalValueType,
        now: datetime.datetime | None = None,
        verified: bool = False
    ) -> Principal:
        """Create a new :class:`~oauthx.types.IPrincipal` instance."""
        mask = value.mask(subject.mask)
        now = now or datetime.datetime.now(datetime.timezone.utc)
        return Principal.model_validate({
            'issuer': issuer,
            'masked': await mask,
            'owner': int(owner),
            'registered': now,
            'value': value,
            'verified': verified
        })

    async def receipt(
        self,
        provider: str | OIDCIssuerIdentifier,
        purpose: Literal['IDENTIFY', 'INVITE', 'LOGIN', 'VERIFY_ACCOUNT'],
        sub: int,
        claims: set[str],
        now: datetime.datetime | None = None,
        request_id: int | None = None,
        client_id: str | None = None
    ):        
        now = now or datetime.datetime.now(datetime.timezone.utc)
        receipt_id = await self.storage.allocate_identifier('Receipt')
        return Receipt.model_validate({
            'id': receipt_id,
            'obtained': now,
            'purpose': purpose,
            'provider': provider,
            'sub': sub,
            'received': claims,
            'client_id': client_id,
            'request_id': request_id
        })

    async def request(
        self,
        client: BaseClient,
        request: fastapi.Request,
        redirect_uri: RedirectURI | None
    ) -> AuthorizationRequest:
        return AuthorizationRequest.model_validate({
            **request.query_params,
            'iss': self.issuer,
            'id': await self.storage.allocate_identifier(Authorization),
            'client_id': str(client.client_id),
            'client_name': client.get_display_name(),
            'redirect_uri': redirect_uri
        })

    async def response_mode(
        self,
        iss: str,
        client: BaseClient,
        response_type: ResponseType | None,
        response_mode: str | None,
        redirect_uri: RedirectURI | None,
        state: str | None
    ) -> IResponseMode:
        return ResponseMode.model_validate({
            'iss': iss,
            'client': client,
            'redirect_uri': redirect_uri,
            'response_type': response_type,
            'response_mode': response_mode,
            'state': state
        })

    async def subject(self, use: Literal['personal', 'institutional'], pk: int | None = None) -> ISubject:
        """Create a new :class:`~oauthx.types.ISubject` instance."""
        k = secrets.token_bytes(32)
        m = secrets.token_bytes(32)
        n = datetime.datetime.now(datetime.timezone.utc)
        return Subject.model_validate({
            'use': use,
            'sub': pk or await self.storage.allocate_identifier(Subject),
            'kind': 'User',
            'dmk': {
                'kty': 'oct',
                'alg': 'HS384',
                'iat': n,
                'kid': b64encode(hashlib.sha256(m).digest(), encoder=bytes.decode),
                'k': b64encode(k, encoder=bytes.decode),
                'use': 'sig',
            },
            'dek': {
                'kty': 'oct',
                'alg': 'dir',
                'iat': n,
                'kid': b64encode(hashlib.sha256(k).digest(), encoder=bytes.decode),
                'use': 'enc',
                'k': b64encode(k, encoder=bytes.decode)
            }
        })

    async def userinfo(
        self,
        client: IClient,
        subject: ISubject,
        scope: set[str],
        session: Session | None = None,
        contributors: list[IUserInfoContributor] | None = None
    ) -> OIDCToken:
        """Construct an :class:`~aiopki.ext.jose.OIDCToken` instance representing
        the claims that are granted under the given `scope`.
        """
        userinfo = await UserInfo.restore(subject, self.storage)
        client.contribute_to_userinfo(userinfo)
        subject.contribute_to_userinfo(userinfo)
        if session is not None:
            session.contribute_to_userinfo(userinfo)
        for contributor in (contributors or []):
            contributor.contribute_to_userinfo(userinfo)
        granted = self._get_granted_claims(scope)
        return OIDCToken.new(120, **{
            **userinfo.dump(granted),
            'aud': {str(client.id)},
            'sub': str(subject.get_primary_key()),
            'iss': self.issuer
        })

    def _get_granted_claims(self, scope: set[str]) -> set[str]:
        claims = [self.scopes[x].claims for x in scope if x in self.scopes]
        return set(functools.reduce(operator.add, claims, []))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import contextlib
import inspect
import logging
import uuid
from typing import Any
from typing import Sequence
from typing import TypeVar

import fastapi
import fastapi.params

from oauthx.lib import utils
from oauthx.server.models import ClientKey
from oauthx.server.request import Request
from ..config import Config
from ..params import AuthorizationServer
from ..params import CurrentConfig
from ..params import Storage


T = TypeVar('T', bound='BaseRequestHandler')


class BaseRequestHandler:
    """Handles a request to the authorization server."""
    __module__: str = 'oauthx.server.handlers'
    dependencies: list[fastapi.params.Depends] = []
    include_in_schema: bool = True
    methods: list[str] = ['GET']
    name: str
    logger: logging.Logger = logging.getLogger('uvicorn')
    path: str
    responses: dict[int | str, Any] = {}
    response_class: type[fastapi.Response] = fastapi.responses.Response
    response_description: str
    route_class: type[fastapi.routing.APIRoute] | None = None
    storage: Storage
    status_code: int = 200
    summary: str
    tags: Sequence[str] = ['OAuth 2.x/OpenID Connect']

    @utils.class_property
    def __signature__(cls) -> inspect.Signature:
        return utils.merge_signatures([
            inspect.signature(cls.__init__),
            inspect.signature(cls.setup)
        ])

    @classmethod
    def inject(cls: type[T]) -> fastapi.Response:
        async def f(request: Request, handler: T = fastapi.Depends(cls)):
            async with handler.transaction(request):
                response = await handler.prepare(request) or await handler.handle(request)
                return await handler.process_response(request, response)
        return fastapi.Depends(f)

    @classmethod
    def add_to_router(
        cls,
        router: fastapi.APIRouter | fastapi.FastAPI,
        name: str | None = None,
        dependencies: Sequence[fastapi.params.Depends] = [],
        route_class: type[fastapi.routing.APIRoute] | None = None,
        path: str | None = None,
        prefix: str = '',
        **kwargs: Any,
    ) -> None:
        async def f(response: Any = cls.inject()):
            return response
        f.__doc__ = cls.__doc__

        if isinstance(router, fastapi.APIRouter):
            route_class = route_class or cls.route_class
            if route_class is not None:
                kwargs['route_class_override'] = route_class
        router.add_api_route(
            methods=cls.methods,
            path=path or f'{prefix}{cls.path}',
            endpoint=f,
            include_in_schema=cls.include_in_schema,
            name=name or cls.name,
            summary=cls.summary,
            description=cls.__doc__,
            status_code=cls.status_code,
            dependencies=[*list(dependencies), *cls.dependencies],
            response_class=cls.response_class,
            response_description=cls.response_description,
            response_model=None,
            responses=dict(cls.responses),
            tags=list(cls.tags),
            **kwargs
        )


    def __init__(
        self,
        server: AuthorizationServer,
        request: Request,
        storage: Storage,
        config: Config = CurrentConfig,
        **kwargs: Any
    ):
        self.config = config
        self.request = request
        self.server = server
        self.storage = storage
        self.setup(**kwargs)

    def redirect(self, url: str, status_code: int = 302) -> fastapi.Response:
        return fastapi.responses.RedirectResponse(
            status_code=status_code,
            url=url
        )

    def setup(self, **_: Any):
        pass

    async def handle(self, request: Request) -> fastapi.Response:
        raise NotImplementedError

    async def incident(self, request: Request) -> fastapi.Response:
        return await request.render_to_response(
            'oauthx/security/incident.html.j2',
            context={
            'incident_id': uuid.uuid4(),
            },
            status_code=403
        )

    @contextlib.asynccontextmanager
    async def transaction(self, request: Request):
        await self.begin(request)
        try:
            yield self
        except Exception as e:
            await self.on_exception(e)
            raise
        else:
            await self.commit(request)

    async def begin(self, request: Request) -> None:
        pass

    async def prepare(self, request: Request) -> fastapi.Response | None:
        pass

    async def process_response(
        self,
        request: Request,
        response: fastapi.Response
    ) -> fastapi.Response:
        return response

    async def commit(self, request: Request) -> None:
        pass

    async def get_client_id(self) -> ClientKey:
        raise NotImplementedError

    async def on_exception(self, exception: Exception) -> None:
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from oauthx.lib.exceptions import InvalidRequest
from oauthx.server.params import CurrentSubject
from oauthx.server.params import ObjectFactory
from oauthx.server.params import PluginRunner
from oauthx.server.params import RequestSession
from oauthx.server.params import TokenIssuer
from oauthx.server.params import TokenSigner
from oauthx.server.request import Request
from oauthx.server.types import LoginRequired
from oauthx.server.types import InvalidResponseType
from oauthx.server.types import UnauthorizedClient
from ..baserequesthandler import BaseRequestHandler
from .params import Client
from .params import AuthorizationRequest
from .params import RedirectURI
from .params import ResponseMode
from .params import ResponseType
from .params import ResourceOwner
from .params import Scope
from .params import State


class AuthorizationRequestHandler(BaseRequestHandler):
    """Provides an interface for the resource owner to authorize a certain
    scope for a client, and redirect back to the clients' redirection
    endpoint.
    """
    __module__: str = 'oauthx.server.handlers'
    client: Client
    name: str = 'oauth2.authorize'
    path: str = '/authorize'
    redirect_uri: RedirectURI
    responses: dict[int | str, Any] = {
        400: {
            'description': (
                "Unrecoverable error that is not allowed to redirect"
            )
        }
    }
    response_class: type[fastapi.Response] = fastapi.responses.RedirectResponse
    response_description: str = "Redirect to the clients' redirection endpoint."
    response_type: ResponseType
    scope: Scope
    status_code: int = 302
    subject: CurrentSubject | None
    summary: str = "Authorization Endpoint"

    def setup(
        self,
        *,
        issuer: TokenIssuer,
        client: Client,
        redirect_uri: RedirectURI,
        subject: CurrentSubject,
        params: AuthorizationRequest,
        plugins: PluginRunner,
        owner: ResourceOwner,
        response_mode: ResponseMode,
        response_type: ResponseType,
        scope: Scope,
        session: RequestSession,
        signer: TokenSigner,
        state: State,
        factory: ObjectFactory,
        **_: Any
    ):
        self.client = client
        self.factory = factory
        self.issuer = issuer
        self.owner = owner
        self.params = params
        self.plugins = plugins
        self.redirect_uri = redirect_uri
        self.response_mode = response_mode
        self.response_type = response_type
        self.scope = scope
        self.session = session
        self.signer = signer
        self.state = state
        self.subject = subject

    async def handle(self, request: Request) -> fastapi.Response:

        if not self.client.allows_response_type(self.response_type):
            raise InvalidResponseType

        if not self.client.can_grant(self.response_mode.grants()):
            self.logger.debug(
                "Client does not allow the requested grant (client: %s)",
                self.client.id
            )
            raise UnauthorizedClient

        if not self.state and self.client.requires_state():
            self.logger.debug(
                "Client requires the state parameter (client: %s)",
                self.client.id
            )
            raise InvalidRequest(
                "The client requires the use of the state "
                "parameter."
            )

        # If the authorization request was not pushed, persist
        # the parameters.
        if self.params is None:
            self.params = await self.factory.request(
                client=self.client, # type: ignore
                request=self.request,
                redirect_uri=self.redirect_uri
            )
            self.logger.debug(
                "Persisted authorization request parameters (client: %s, urn: %s)",
                self.client.id, self.params.request_uri
            )
        await self.storage.persist(self.params)
        if self.subject is None:
            self.logger.debug(
                "Request is not authenticated (client: %s, request: %s)",
                self.client.id, self.params.request_uri
            )
            raise LoginRequired(
                client=self.client,
                next_url=self.params.get_authorize_url(request),
                deny_url=await self.response_mode.deny(),
                a=self.params.id
            )
        if self.owner is None:
            raise NotImplementedError

        response = await self.plugins.validate_scope(
            client=self.client,
            request=self.params,
            scope=self.params.scope
        )
        if response is not None:
            return response
        assert self.params.id
        authorization = await self.factory.authorization(
            request=self.params,
            client_id=self.client.id,
            lifecycle='GRANTED',
            scope=self.scope,
            sub=self.subject.get_primary_key(), # type: ignore
            token_types=self.response_mode.get_token_types()
        )
        authorization.contribute(self.params)
        authorization.contribute(self.session)
        await self.storage.persist(authorization)
        await self.storage.delete(self.params)
        
        code = await self.issuer.authorization_code(
            signer=self.signer,
            client=self.client,
            owner=self.owner,
            authorization_id=authorization.id,
            sub=self.subject.get_primary_key(), # type: ignore
            redirect_uri=self.params.redirect_uri
        )
        return await self.response_mode.redirect(
            signer=self.signer,
            code=code
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import logging
import uuid
from typing import Literal

import fastapi

from oauthx.lib.types import OIDCIssuerIdentifier
from .protocols import ISubjectLogger


class SubjectLogger(ISubjectLogger):
    __module__: str = 'oauthx.server'
    logger: logging.Logger
    request: fastapi.Request

    def __init__(self, request: fastapi.Request):
        self.logger = logging.getLogger('oauthx.events.subject')
        self.request = request
        setattr(request, 'subject_logger', self)

    def onboarded(
        self,
        sub: int,
        registrar: str,
        authorization_id: int | None = None,
        timestamp: datetime.datetime | None = None,
        client_id: str | None = None
    ) -> None:
        dto = {
            'apiVersion': 'v1',
            'kind': 'SubjectOnboarded',
            'metadata': {
                'id': str(uuid.uuid4()),
                'name': f'onboarded/{sub}',
                'host': self.get_host(),
                'timestamp': (timestamp or self.now()).isoformat(timespec='seconds'),
                'endpoint': self.request.scope['route'].name
            },
            'data': {
                'sub': sub,
                'authorizationId': authorization_id,
                'clientId': client_id,
                'registrar': registrar
            }
        }
        self.logger.info(dto)

    def receipt(
        self,
        receipt_id: int,
        sub: int,
        obtained: datetime.datetime,
        processed: set[str],
        provider: OIDCIssuerIdentifier,
        purpose: Literal['IDENTIFY', 'INVITE', 'LOGIN', 'VERIFY_ACCOUNT'],
        received: set[str],
        client_id: str | None = None,
        request_id: int | None = None,
        timestamp: datetime.datetime | None = None
    ):
        dto = {
            'apiVersion': 'v1',
            'kind': 'PersonalDataReceived',
            'metadata': {
                'id': str(uuid.uuid4()),
                'name': f'receipts/{receipt_id}',
                'host': self.get_host(),
                'timestamp': (timestamp or self.now()).isoformat(timespec='seconds'),
                'endpoint': self.request.scope['route'].name
            },
            'data': {
                'authorizationId': request_id,
                'clientId': client_id,
                'receiptId': receipt_id,
                'sub': sub,
                'obtained': obtained.isoformat(timespec='seconds'),
                'processed': list(processed),
                'provider': provider,
                'purpose': purpose,
                'received': list(received)
            }
        }
        self.logger.info(dto)

    def get_host(self) -> str:
        assert self.request.client is not None
        return self.request.client.host

    def now(self) -> datetime.datetime:
        return datetime.datetime.now(datetime.timezone.utc)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import Literal

import pydantic


class IssuedToken(pydantic.BaseModel):
    active: bool
    aud: str | list[str] = []
    client_id: str | None = None
    exp: int
    iat: int
    iss: str
    jti: str
    nbf: int
    scope: set[str] | None = None
    sub: str | None = None
    token_type: Literal['Bearer'] = 'Bearer'

    @classmethod
    def new(cls, ttl: int, **kwargs: Any):
        now = datetime.datetime.now(datetime.timezone.utc)
        params = {
            **kwargs,
            'active': True,
            'exp': int((now + datetime.timedelta(seconds=ttl)).timestamp()),
            'iat': int(now.timestamp()),
            'nbf': int(now.timestamp()),
        }

        return cls.model_validate(params)

    @pydantic.field_serializer('scope')
    def serialize_scope(self, value: set[str], _: Any) -> str | None:
        return str.join(' ', sorted(self.scope)) if self.scope else None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi
import fastapi.responses

from oauthx.lib.protocols import IClient


class LoginRequired(RuntimeError):
    __module__: str = 'oauthx.types'
    client: IClient

    @property
    def client_id(self) -> str:
        return str(self.client.id)

    def __init__(self, client: IClient, deny_url: str | None = None, next_url: str | None = None, **params: Any):
        self.client = client
        self.deny_url = deny_url
        self.next_url = next_url
        self.params = params

    def redirect(self, request: fastapi.Request, name: str, next_url: str) -> fastapi.Response:
        params: dict[str, str] = {
            **self.params,
            'client_id': self.client.id
        }
        if self.deny_url:
            params['d'] = self.deny_url
        if next_url:
            params['n'] = self.next_url or next_url
        return fastapi.responses.RedirectResponse(
            status_code=302,
            url=request.url_for(name).include_query_params(**params),
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from oauthx.lib.protocols import IClient
from oauthx.server import Plugin
from oauthx.server.models import Authorization
from oauthx.server.models import Principal
from oauthx.server.params import PendingAuthorizationRequest
from oauthx.server.params import SubjectLogger
from oauthx.server.types import IAuthorizationRouter
from oauthx.server.types import StopSnooping
from .const import ISSUER
from .params import TelegramBotName
from .params import TelegramUserInfo
from .types import TelegramAccountUnusable


class TelegramPlugin(Plugin):
    __module__: str = 'oauthx.ext.telegram'
    
    @classmethod
    def __register__(cls, router: IAuthorizationRouter) -> None:
        router.add_plugin(
            methods=['GET'],
            path='telegram/confirm',
            handler=cls,
            method=cls.login,
            name='oauth.ext.telegram.login',
            description=(
                "Login with Telegram to register an account."
            ),
            authenticated=True
        )
        router.add_plugin(
            methods=['GET'],
            path='telegram/welcome',
            handler=cls,
            method=cls.welcome,
            name='oauth.ext.telegram.welcome',
            description="Callback for Telegram post-login.",
            authenticated=True,
            needs_data=True
        )
        router.register_template_module('oauthx.ext.telegram')

    async def welcome(
        self,
        authnreq: PendingAuthorizationRequest,
        userinfo: TelegramUserInfo,
        bot_name: TelegramBotName,
        logger: SubjectLogger
    ) -> fastapi.Response:
        # Retrieve or create a principal and determine if the currently
        # authenticated user has permission to use it.
        assert self.subject is not None
        context = {
            'telegram_bot_name': bot_name,
            'telegram_callback_url': self.request.url_for(
                'oauth.ext.telegram.welcome',
                pk=str(authnreq.pk)
            ),
            'telegram_username':  userinfo.username
        }
        principal = await self.storage.get(userinfo.masked)
        sub = int(self.subject.get_primary_key()) # type: ignore
        if principal is None:
            principal = await self.factory.principal(
                subject=self.subject,
                issuer=userinfo.issuer,
                owner=self.subject.get_primary_key(), # type: ignore
                verified=True,
                value=userinfo.id
            )
            await principal.encrypt(self.subject)
            await self.storage.persist(principal)
            
            receipt = await self.factory.receipt(
                provider=userinfo.issuer,
                purpose='VERIFY_ACCOUNT',
                sub=sub,
                claims={
                    'telegram',
                    'given_name',
                    'preferred_username',
                    'family_name'
                },
                client_id=str(authnreq.client_id),
                request_id=authnreq.id
            )
            async with receipt.transaction(self.subject, self.storage, logger) as tx:
                tx.add(
                    kind='telegram',
                    value=str(userinfo.id.sub),
                    issuer=userinfo.issuer
                )
        elif not principal.is_owned_by(sub):
            raise TelegramAccountUnusable(context)
        authnreq.add_to_template_context(self.request, context)
        return await self.request.render_to_response(
            template_names=f'oauthx/telegram/welcome.html.j2',
            context=context
        )

    async def login(
        self,
        authnreq: PendingAuthorizationRequest,
        bot_name: TelegramBotName,
    ) -> fastapi.Response:
        if self.subject is None:
            raise StopSnooping
        context = {
            'telegram_bot_name': bot_name,
            'telegram_callback_url': self.request.url_for(
                'oauth.ext.telegram.welcome',
                pk=str(authnreq.pk)
            ),
        }
        authnreq.add_to_template_context(self.request, context)
        return await self.request.render_to_response(
            template_names=f'oauthx/telegram/login.html.j2',
            context=context
        )

    def handles_scope(self, name: str) -> bool:
        return name == 'telegram'

    async def resolve_scope(
        self,
        client: IClient,
        authorization: Authorization,
        name: str
    ) -> tuple[str | None, dict[str, str]]:
        assert self.subject is not None
        principals = [x async for x in self.storage.find(
            kind=Principal,
            filters=[
                ('kind', '=', 'subject'),
                ('issuer', '=', ISSUER),
                ('owner', '=', int(self.subject.get_primary_key())) # type: ignore
            ]
        )]
        if principals:
            return None, {}

        return 'oauth.ext.telegram.login', {}


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
import inspect
import logging
from typing import Annotated
from typing import Any
from typing import Iterable
from typing import TypeAlias

import fastapi
from aiopki.ext import jose
from aiopki.ext.jose import OIDCToken
from aiopki.types import ISigner
from aiopki.utils import b64encode

from oauthx.lib import utils
from oauthx.lib import RFC9068AccessToken
from oauthx.lib.protocols import IClient
from oauthx.lib.types import AccessToken
from oauthx.lib.types import RedirectURI
from oauthx.server.config import Config
from oauthx.server.models import SubjectKey
from oauthx.server.protocols import IResourceOwner
from oauthx.server.request import Request
from oauthx.server.types import TokenMAC
from .currentconfig import CurrentConfig
from .datamaskingkey import DataMaskingKey
from .issuer import Issuer
from .storage import Storage


__all__: list[str] = ['TokenIssuer']


class _TokenIssuer:
    __module__: str = 'oauthx.server.handlers'
    default_ttl: int = 3600
    logger: logging.Logger = logging.getLogger('uvicorn')
    storage: Storage

    @utils.class_property
    def __signature__(cls) -> inspect.Signature:
        return utils.merge_signatures([
            inspect.signature(cls.__init__),
            inspect.signature(cls.setup)
        ])

    def __init__(
        self,
        request: Request,
        issuer: Issuer,
        masking_key: DataMaskingKey,
        storage: Storage,
        config: Config = CurrentConfig,
        **kwargs: Any
    ):
        self.config = config
        self.issuer = issuer
        self.masking_key = masking_key
        self.request = request
        self.storage = storage
        self.setup(**kwargs)

    async def access_token(
        self,
        signer: ISigner,
        client_id: str,
        sub: str,
        aud: str | Iterable[str] | None = None,
        scope: set[str] | None = None,
        **claims: Any
    ) -> AccessToken:
        token = RFC9068AccessToken.new(
            ttl=self.default_ttl,
            client_id=client_id,
            aud=aud or client_id,
            iss=self.issuer,
            scope=scope or set(),
            sub=sub,
            **claims
        )
        return AccessToken(await token.sign(signer))

    async def authorization_code(
        self,
        signer: ISigner,
        client: IClient,
        owner: IResourceOwner,
        authorization_id: int,
        sub: SubjectKey,
        redirect_uri: RedirectURI | str | None
    ) -> str:
        mac = TokenMAC()
        mac.add(client)
        mac.add(owner)
        params: dict[str, Any] = {
            'aut': int(authorization_id),
            'client_id': str(client.id),
            'sub': str(sub),
            'mac': str(mac),
        }
        if redirect_uri is not None:
            params['redirect_uri'] = b64encode(
                hashlib.sha3_256(str.encode(redirect_uri)).digest(),
                encoder=bytes.decode
            )
        jws = jose.jws(params)
        await jws.sign(signer.default_algorithm(), signer, {'typ': 'jwt+code'})
        return jws.encode(encoder=bytes.decode)

    async def id_token(
        self,
        signer: ISigner,
        userinfo: OIDCToken,
        access_token: AccessToken | None = None,
        code: str | None = None,
    ) -> jose.JOSEObject:
        jws = jose.jws(userinfo.model_dump(), {'typ': 'jwt'})
        await jws.sign(signer.default_algorithm(), signer)
        return jws

    def setup(self, **kwargs: Any):
        pass


TokenIssuer: TypeAlias = Annotated[_TokenIssuer, fastapi.Depends(_TokenIssuer)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import MutableMapping

import fastapi
import pydantic
from aiopki.utils import b64encode_int
from canonical.exceptions import ProgrammingError

from oauthx.lib.protocols import IClient
from oauthx.lib.protocols import IStorage
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ResponseType
from oauthx.lib.types import ScopeType
from oauthx.server.types import AuthorizationRequestKey
from .authorizationcontext import AuthorizationContext
from .clientkey import ClientKey
from .responsemode import ResponseMode


class AuthorizationRequest(pydantic.BaseModel):
    id: int
    client_name: str
    iss: str
    
    # https://datatracker.ietf.org/doc/html/rfc6749
    client_id: ClientKey = pydantic.Field(
        default=...
    )

    redirect_uri: RedirectURI | None = pydantic.Field(
        default=None
    )

    response_type: ResponseType = pydantic.Field(
        default=...
    )

    scope: ScopeType = pydantic.Field(
        default_factory=ScopeType
    )

    state: str | None = pydantic.Field(
        default=None
    )
    
    # https://openid.net/specs/openid-connect-core-1_0.html
    nonce: str | None = pydantic.Field(
        default=None
    )

    # https://openid.net/specs/oauth-v2-multiple-response-types-1_0.html
    response_mode: str | None = pydantic.Field(
        default=None
    )

    @property
    def request_uri(self) -> str:
        return f'urn:ietf:params:oauth:request_uri:{b64encode_int(self.id, encoder=bytes.decode)}'

    @property
    def pk(self) -> AuthorizationRequestKey:
        return AuthorizationRequestKey(self.id)

    def add_to_template_context(self, request: fastapi.Request, context: dict[str, Any]):
        context.update({
            'authorize_url': request.url_for('oauth2.authorize').include_query_params(
                client_id=self.client_id,
                request_uri=self.request_uri
            ),
            'client_name': self.client_name,
            'deny_url': request.url_for('oauth2.deny', pk=self.id),
        })

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        if self.nonce is not None:
            userinfo['nonce'] = self.nonce

    def get_authorize_url(self, request: fastapi.Request) -> str:
        url = request.url_for('oauth2.authorize')\
            .include_query_params(
                client_id=self.client_id,
                request_uri=self.request_uri
            )
        return str(url)

    def get_response_mode(self, client: IClient) -> ResponseMode:
        return ResponseMode.model_validate({
            'iss': self.iss,
            'client': client,
            'redirect_uri': self.redirect_uri,
            'state': self.state,
            'response_type': self.response_type,
            'response_mode': self.response_mode
        })

    async def deny(self, client: IClient, exception: bool = False) -> str:
        response_mode = self.get_response_mode(client)
        return await response_mode.deny()

    async def get_client(self, storage: IStorage) -> IClient:
        client = await storage.get(self.client_id)
        if client is None:
            raise ProgrammingError(
                "The authorization request references a client that does "
                "not exist."
            )
        return client

    async def get_context(self, storage: IStorage) -> AuthorizationContext:
        client = await self.get_client(storage)
        return AuthorizationContext.model_validate({
            'client': client.root, # type: ignore
            'response_mode': self.get_response_mode(client.root) # type: ignore
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .models import AuthorizationRequest
from .request import Request


class Policy:
    __module__: str = 'oauthx.server'

    async def apply(
        self,
        request: Request,
        params: AuthorizationRequest
    ) -> None:
        pass

    def error(
        self,
        error: str,
        error_description: str | None = None,
        error_uri: str | None = None
    ) -> None:
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from aiopki.ext.jose import AudienceType
from aiopki.ext.jose import CompactSerialized

from oauthx.lib.types import GrantType
from oauthx.server.protocols import IResourceOwner
from .clientkey import ClientKey


class Assertion(pydantic.BaseModel):
    grant_type: GrantType
    jws: CompactSerialized
    aud: AudienceType
    iss: str
    sub: str

    @property
    def client_id(self) -> ClientKey:
        return ClientKey(self.iss)

    @property
    def owner(self) -> IResourceOwner.Key:
        return IResourceOwner.Key(self.iss, self.sub)

    def is_client(self) -> bool:
        """Return a boolean if the assertion was created by a
        client. This is detected based on the `iss` claim, as
        a client identifier should not be a URL.
        """
        return self.iss != self.sub

    def is_self_issued(self) -> bool:
        """Return a boolean indicating if the assertion is
        self-issued.
        """
        return self.iss == self.sub


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import functools
from typing import Annotated
from typing import Any
from typing import Callable
from typing import TypeAlias
from typing import TypeVar

import fastapi
from aiopki.types import SubjectID
from canonical import EmailAddress
from canonical import Phonenumber

from oauthx.lib.types import MaskedPrincipal
from oauthx.server.models import PrincipalValueType
from .datamaskingkey import DataMaskingKey


T = TypeVar('T')


class _Masker:
    __module__: str = 'oauthx.types'
    
    def __init__(self, key: DataMaskingKey):
        self.key = key

    @functools.singledispatchmethod
    async def mask(
        self,
        value: PrincipalValueType | str,
        factory: Callable[[str], T] = MaskedPrincipal
    ) -> T:
        if isinstance(value, str):
            sig = await self.key.sign(str.encode(value))
            return factory(bytes.decode(base64.urlsafe_b64encode(sig)))
        raise NotImplementedError(type(value).__name__)

    @mask.register
    async def _(self, value: EmailAddress, factory: Any = MaskedPrincipal) -> Any:
        return await self.mask(f'email:{str(str.lower(value))}', factory=factory)

    @mask.register
    async def _(self, value: Phonenumber, factory: Any = MaskedPrincipal) -> Any:
        return await self.mask(f'phone:{str(str.lower(value))}', factory=factory)

    @mask.register
    async def _(self, value: SubjectID, factory: Any = MaskedPrincipal) -> Any:
        return await self.mask(f'sub:{value.iss}/{value.sub}', factory=factory)


Masker: TypeAlias = Annotated[_Masker, fastapi.Depends(_Masker)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.server.config import Config
from .currentconfig import CurrentConfig


__all__: list[str] = ['IssuerIdentifier']


def get(config: Config = CurrentConfig) -> str:
    return config.issuer.id



IssuerIdentifier: TypeAlias = Annotated[str, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

from aiopki.types import ISigner
from aiopki.ext import jose
from aiopki.ext.jose import CompactSerialized
from aiopki.ext.jose import JWT

from .baseassertion import BaseAssertion


class JWTBearerAssertion(BaseAssertion):
    grant_type: Literal['urn:ietf:params:oauth:grant-type:jwt-bearer']
    assertion: CompactSerialized

    @classmethod
    async def new(
        cls,
        signer: ISigner,
        iss: str,
        sub: str,
        token_endpoint: str,
        ttl: int = 15,
        **claims: Any
    ) -> 'JWTBearerAssertion':
        jwt = JWT.new(
            ttl=ttl,
            aud=token_endpoint,
            iss=iss,
            sub=sub,
            **claims
        )
        jws = jose.jws(jwt.model_dump())
        await jws.sign(signer.default_algorithm(), signer)
        return cls.model_validate({
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
            'assertion': jws
        })

    def must_authenticate(self) -> bool:
        return False

    def must_identify(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

from canonical import ResourceIdentifier


class IRequestSession(Protocol):
    __module__: str = 'oauthx.types'

    class Key(ResourceIdentifier[str, 'IRequestSession']):
        session_id: str
        openapi_example: str = 'client-123'
        openapi_title: str = 'Session ID'

        def __init__(self, session_id: str):
            self.session_id = session_id

        def __str__(self) -> str:
            return self.session_id
        
        def __eq__(self, key: object) -> bool:
            return isinstance(key, type(self)) and key.session_id == self.session_id

    def authenticate(self, sub: str, email: str | None) -> None:
        """Authenticates the session as the given subject."""
        ...

    def claims(self) -> dict[str, Any]:
        """Return a dictionary holding the claims in the session."""
        ...

    def is_dirty(self) -> bool:
        """Return a boolean indicating if the session is dirty."""
        ...

    def set(self, key: str, value: Any) -> None:
        """Set a key to the session with the given value, and mark the
        session as dirty.
        """
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from oauthx.server.config import ProviderConfig
from oauthx.server.handlers import BaseLoginRequestHandler


class LoginRequestHandler(BaseLoginRequestHandler):
    __module__: str = 'oauthx.ext.google'

    def get_context(self, request: fastapi.Request) -> dict[str, Any]:
        ctx = super().get_context(request)
        return {
            **ctx,
            'login_internal_url': request.url_for('oauth2.login.enrolled')\
                .include_query_params(**self.request.query_params)
        }

    def get_providers(self) -> list[ProviderConfig]:
        return [x for x in self.config.providers if x.audience == 'personal']


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


__all__: list[str] = [
    'SectorIdentifierURI'
]


class SectorIdentifierURI(str):
    _sector: str

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.chain_schema([
                core_schema.is_instance_schema(str),
                core_schema.no_info_plain_validator_function(cls),
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @property
    def sector(self) -> str:
        return self._sector

    def __new__(cls, value: str) -> 'SectorIdentifierURI':
        p = urllib.parse.urlparse(value)
        if p.scheme != 'https':
            raise ValueError("A Sector Identifier URI must use the https scheme.")
        if p.hostname is None:
            raise ValueError("A Sector Identifier URI point to a valid host.")
        self = super().__new__(cls, value)
        self._sector = p.hostname
        return self


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pytest

import oauthx


@pytest.mark.skip("Not implemented")
@pytest.mark.asyncio
@pytest.mark.parametrize("redirect_uri", [
    "http://localhost", # May not be localhost
    "http://example.com", # Must use TLS
])
async def test_invalid_redirect_uri_does_not_redirect(
    http: httpx.AsyncClient,
    client: oauthx.Client,
    redirect_uri: str,
):
    response = await http.get(
        url='https://127.0.0.1/oauth/v2/authorize',
        params={
            'client_id': client.client_id,
            'response_type': 'code',
            'redirect_uri': redirect_uri
        }
    )
    assert response.status_code == 403


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

import pytest

import oauthx
from .types import AuthorizeFactory


@pytest.mark.asyncio
async def test_client_id_must_match(
    client: oauthx.Client,
    authorize_request: AuthorizeFactory
):
    client.client_id = bytes.hex(os.urandom(16))
    await authorize_request(
        response_type='code',
        status_code=400,
        expect_media_type='text/html; charset=utf-8'
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib

from aiopki.utils import b64encode

from .ikeypart import IKeyPart

class TokenMAC:
    __module__: str = 'oauthx.server.types'

    def __init__(self):
        self.hasher = hashlib.sha256()

    def add(self, part: IKeyPart) -> None:
        part.update_key(self.hasher.update)

    def __str__(self) -> str:
        return bytes.decode(bytes(self))

    def __bytes__(self) -> bytes:
        return b64encode(self.hasher.digest())


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import secrets
import urllib.parse
from typing import Any
from typing import Callable
from typing import Literal

import pydantic
from aiopki.types import Base64
from aiopki.ext.jose import JOSEObject
from aiopki.ext.jose import JWA
from canonical.security import DomainACL
from fastapi.security import HTTPBasicCredentials

from oauthx.lib.types import GrantType
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ResponseType
from oauthx.lib.types import SectorIdentifierURI
from oauthx.server.protocols import IRequestSubject
from oauthx.server.protocols import IResourceOwner
from oauthx.server.types import MissingRedirectURI
from oauthx.lib  import ClientAuthorizationState
from oauthx.lib  import Provider


class BaseClient(pydantic.BaseModel):
    client_id: str
    client_name: str
    grant_types: list[str] = []
    key: Base64 = pydantic.Field(default_factory=lambda: secrets.token_bytes(32))
    redirect_uris: list[RedirectURI] = []
    response_types: list[ResponseType] = []
    scope: set[str] = set()
    require_pushed_authorization_requests: bool = False
    token_endpoint_auth_method: Any = None

    # OpenID Connect Dynamic Client Registration 1.0 incorporating errata set 2
    sector_identifier_uri: SectorIdentifierURI | None = None
    subject_type: Literal['public', 'pairwise'] = 'pairwise'
    
    # JWT Secured Authorization Response Mode for OAuth 2.0 (JARM)
    authorization_signed_response_alg: JWA = pydantic.Field(
        default=JWA.rs256
    )

    # Custom properties
    domains: DomainACL = pydantic.Field(
        default_factory=DomainACL.null
    )

    @property
    def id(self) -> str:
        return self.client_id

    @property
    def provider(self) -> Provider:
        raise NotImplementedError

    @property
    def sector_identifier(self) -> str:
        hostnames = {p.hostname or '' for p in map(urllib.parse.urlparse, self.redirect_uris)}
        if (not hostnames or len(hostnames) > 1) and not self.sector_identifier_uri:
            raise NotImplementedError
        if self.sector_identifier_uri:
            return self.sector_identifier_uri.sector
        else:
            return hostnames.pop()

    def allows_response_type(self, response_type: ResponseType) -> bool:
        return response_type in self.response_types

    def model_post_init(self, _: Any) -> None:
        if ('authorization_code' in self.grant_types) and not self.redirect_uris:
            raise ValueError("At least one redirect URI must be specified.")

    @functools.singledispatchmethod
    async def authenticate(self, credential: JOSEObject | HTTPBasicCredentials | str | None) -> bool:
        return False

    async def authorize(self, **kwargs: Any) -> ClientAuthorizationState:
        raise NotImplementedError

    async def authorization_code(self, **kwargs: Any) -> None:
        raise NotImplementedError

    def allows_scope(self, scope: set[str]) -> bool:
        return scope <= self.scope

    def can_grant(self, grant_type: GrantType | None) -> bool:
        """Return a boolean indicating if the client allows the given
        `grant_type`.
        """
        return any([
            grant_type is None,
            grant_type in self.grant_types
        ])

    def can_redirect(self, redirect_uri: RedirectURI | None) -> bool:
        return redirect_uri is None or any([x.can_redirect(redirect_uri) for x in self.redirect_uris])

    def is_confidential(self) -> bool:
        return True

    def get_default_redirect_uri(self) -> RedirectURI:
        if not self.redirect_uris:
            raise MissingRedirectURI
        return self.redirect_uris[0]

    def get_display_name(self) -> str:
        return self.client_name

    def get_sector_identifier(self) -> str:
        return self.sector_identifier

    def has_redirection_endpoints(self) -> bool:
        return len(self.redirect_uris) > 0

    def must_push(self) -> bool:
        return self.require_pushed_authorization_requests

    def requires_state(self) -> bool:
        return False

    def resource_owner(self, subject: IRequestSubject) -> IResourceOwner.Key:
        return IResourceOwner.Key(self.client_id, str(subject.id)) # type: ignore

    def update_key(self, update: Callable[[bytes], None]) -> None:
        update(self.key)

    def wants_encryption(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json

from aiopki.ext import jose
from aiopki.types import ISigner
from canonical import UnixTimestamp

from oauthx.lib.models import ClientKey
from oauthx.lib.types import ScopeType


class RFC9068AccessToken(jose.JWT):
    acr: str | None = None
    amr: str | None = None
    auth_time: UnixTimestamp | None = None
    client_id: ClientKey
    iss: str # type: ignore
    scope: ScopeType = ScopeType()
    sub: str # type: ignore

    async def sign(self, signer: ISigner) -> str:
        claims = json.loads(self.model_dump_json(exclude_defaults=True, exclude_none=True))
        jws = jose.jws(claims)
        await jws.sign(signer.default_algorithm(), signer, {'typ': 'at+jwt'})
        return jws.encode(encoder=bytes.decode)

    def is_client(self) -> bool:
        """Return a boolean indicating if the access token represents
        an OAuth 2.x/OpenID Connect client.
        """
        return self.client_id == self.sub

    def is_openid(self) -> bool:
        """Return a boolean indicating if the access token was issued
        to an OpenID Connect client.
        """
        return 'openid' in self.scope

    def validate_scope(self, scope: set[str]) -> bool:
        return scope <= self.scope


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .iauthorizationserver import IAuthorizationServer
from .iclient import IClient
from .iplugin import IPlugin
from .irequestsession import IRequestSession
from .irequestsubject import IRequestSubject
from .iresourceowner import IResourceOwner
from .isubject import ISubject
from .isubjectlogger import ISubjectLogger
from .itokenissuer import ITokenIssuer
from .iuserinfocontributor import IUserInfoContributor


__all__: list[str] = [
    'IAuthorizationServer',
    'IClient',
    'IPlugin',
    'IRequestSession',
    'IRequestSubject',
    'IResourceOwner',
    'ISubject',
    'ISubjectLogger',
    'ITokenIssuer',
    'IUserInfoContributor'
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from .baseassertion import BaseAssertion


class SAML2BearerAssertion(BaseAssertion):
    grant_type: Literal['urn:ietf:params:oauth:grant-type:saml2-bearer']